# This tests calculates the force on the atoms of a small molecule.
#
# If the test fails, set the fd boolean below to enable a (costly) finite
# difference check.

import numpy as np
import pytest
from ase.build import molecule

from gpaw import GPAW
from gpaw.atom.basis import BasisMaker


@pytest.mark.later  # basis set cutoff?
def test_lcao_force():
    obasis = BasisMaker('O').generate(2, 1, energysplit=0.3, tailnorm=0.03**.5)
    hbasis = BasisMaker('H').generate(2, 1, energysplit=0.3, tailnorm=0.03**.5)
    basis = {'O': obasis, 'H': hbasis}

    system = molecule('H2O')
    system.center(vacuum=1.5)
    system.rattle(stdev=.2, seed=42)
    system.set_pbc(1)

    calc = GPAW(h=0.2,
                mode='lcao',
                basis=basis,
                kpts=[(0., 0., 0.), (.3, .1, .4)],
                convergence={'density': 1e-5, 'energy': 1e-6})

    system.calc = calc

    # Previous FD result, generated by disabled code below
    F_ac_ref = np.array([[1.05022478, 1.63103681, -5.00612007],
                         [-0.69739179, -0.89624274, 3.03203147],
                         [-0.34438181, -0.7042696, 1.77209023]])

    for use_rho in [0, 1]:
        if use_rho:
            for kpt in calc.wfs.kpt_u:
                kpt.rho_MM = calc.wfs.calculate_density_matrix(kpt.f_n,
                                                               kpt.C_nM)

        F_ac = system.get_forces()
        system.calc.results.pop('forces')

        err_ac = np.abs(F_ac - F_ac_ref)
        err = err_ac.max()

        print('Force')
        print(F_ac)
        print()
        print('Reference result')
        print(F_ac_ref)
        print()
        print('Error')
        print(err_ac)
        print()
        print('Max error')
        print(err)

        # ASE uses dx = [+|-] 0.001 by default,
        # error should be around 2e-3.
        # In fact 4e-3 would probably be acceptable
        assert err < 3e-3

    # Set boolean to run new FD check
    fd = False

    if fd:
        from ase.calculators.test import numeric_force
        F_ac_fd = np.array([[numeric_force(system, a, i)
                             for i in range(3)]
                            for a in range(len(system))])
        print('Self-consistent forces')
        print(F_ac)
        print('FD')
        print(F_ac_fd)
        print(repr(F_ac_fd))
        print(F_ac - F_ac_fd, np.abs(F_ac - F_ac_fd).max())

        assert np.abs(F_ac - F_ac_fd).max() < 4e-3



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""
Python wrapper functions for the ``C`` package:
Basic Linear Algebra Subroutines (BLAS)

See also:
http://en.wikipedia.org/wiki/Basic_Linear_Algebra_Subprograms
and
http://www.netlib.org/lapack/lug/node145.html
"""
from typing import TypeVar

import numpy as np
import scipy.linalg.blas as blas

import _gpaw
from gpaw import debug
from gpaw.new import prod
from gpaw.typing import Array2D, ArrayND

__all__ = ['mmm']

T = TypeVar('T', float, complex)


def mmm(alpha: T,
        a: Array2D,
        opa: str,
        b: Array2D,
        opb: str,
        beta: T,
        c: Array2D) -> None:
    """Matrix-matrix multiplication using dgemm or zgemm.

    For opa='N' and opb='N', we have:::

        c <- αab + βc.

    Use 'T' to transpose matrices and 'C' to transpose and complex conjugate
    matrices.
    """

    assert opa in 'NTC'
    assert opb in 'NTC'

    if opa == 'N':
        a1, a2 = a.shape
    else:
        a2, a1 = a.shape
    if opb == 'N':
        b1, b2 = b.shape
    else:
        b2, b1 = b.shape
    assert a2 == b1
    assert c.shape == (a1, b2)

    assert a.dtype == b.dtype == c.dtype
    assert a.strides[1] == c.itemsize or a.size == 0
    assert b.strides[1] == c.itemsize or b.size == 0
    assert c.strides[1] == c.itemsize or c.size == 0
    if a.dtype == float:
        assert not isinstance(alpha, complex)
        assert not isinstance(beta, complex)
    else:
        assert a.dtype == complex

    _gpaw.mmm(alpha, a, opa, b, opb, beta, c)


def to2d(array: ArrayND) -> Array2D:
    """2D view af ndarray.

    >>> to2d(np.zeros((2, 3, 4))).shape
    (2, 12)
    """
    shape = array.shape
    return array.reshape((shape[0], prod(shape[1:])))


def mmmx(alpha: T,
         a: ArrayND,
         opa: str,
         b: ArrayND,
         opb: str,
         beta: T,
         c: ArrayND) -> None:
    """Matrix-matrix multiplication using dgemm or zgemm.

    Arrays a, b and c are converted to 2D arrays before calling mmm().
    """
    mmm(alpha, to2d(a), opa, to2d(b), opb, beta, to2d(c))


def axpy(alpha, x, y):
    """alpha x plus y.

    Performs the operation::

      y <- alpha * x + y

    """
    if x.size == 0:
        return
    assert x.flags.contiguous
    assert y.flags.contiguous
    x = x.ravel()
    y = y.ravel()
    if x.dtype == float:
        z = blas.daxpy(x, y, a=alpha)
    else:
        z = blas.zaxpy(x, y, a=alpha)
    assert z is y, (x, y, x.shape, y.shape)


def rk(alpha, a, beta, c, trans='c'):
    """Rank-k update of a matrix.

    For ``trans='c'`` the following operation is performed:::

              †
      c <- αaa + βc,

    and for ``trans='t'`` we get:::

             †
      c <- αa a + βc

    If the ``a`` array has more than 2 dimensions then the 2., 3., ...
    axes are combined.

    Only the lower triangle of ``c`` will contain sensible numbers.
    """
    assert beta == 0.0 or np.isfinite(c).all()

    assert (a.dtype == float and c.dtype == float or
            a.dtype == complex and c.dtype == complex)
    assert a.flags.contiguous
    assert a.ndim > 1
    if trans == 'n':
        assert c.shape == (a.shape[1], a.shape[1])
    else:
        assert c.shape == (a.shape[0], a.shape[0])
    assert c.strides[1] == c.itemsize
    _gpaw.rk(alpha, a, beta, c, trans)


def r2k(alpha, a, b, beta, c, trans='c'):
    """Rank-2k update of a matrix.

    Performs the operation::

                        dag        cc       dag
      c <- alpha * a . b    + alpha  * b . a    + beta * c

    or if trans='n'::
                    dag           cc   dag
      c <- alpha * a   . b + alpha  * b   . a + beta * c

    where ``a.b`` denotes the matrix multiplication defined by::

                 _
                \
      (a.b)   =  ) a         * b
           ij   /_  ipklm...     pjklm...
               pklm...

    ``cc`` denotes complex conjugation.

    ``dag`` denotes the hermitian conjugate (complex conjugation plus a
    swap of axis 0 and 1).

    Only the lower triangle of ``c`` will contain sensible numbers.
    """
    assert beta == 0.0 or np.isfinite(np.tril(c)).all()

    assert (a.dtype == float and b.dtype == float and c.dtype == float or
            a.dtype == complex and b.dtype == complex and c.dtype == complex)
    assert a.flags.contiguous and b.flags.contiguous
    assert a.ndim > 1
    assert a.shape == b.shape
    if trans == 'c':
        assert c.shape == (a.shape[0], a.shape[0])
    else:
        assert c.shape == (a.shape[1], a.shape[1])
    assert c.strides[1] == c.itemsize
    _gpaw.r2k(alpha, a, b, beta, c, trans)


def _gemmdot(a, b, alpha=1.0, beta=1.0, out=None, trans='n'):
    """Matrix multiplication using gemm.

    return reference to out, where::

      out <- alpha * a . b + beta * out

    If out is None, a suitably sized zero array will be created.

    ``a.b`` denotes matrix multiplication, where the product-sum is
    over the last dimension of a, and either
    the first dimension of b (for trans='n'), or
    the last dimension of b (for trans='t' or 'c').

    If trans='c', the complex conjugate of b is used.
    """
    # Store original shapes
    ashape = a.shape
    bshape = b.shape

    # Vector-vector multiplication is handled by dotu
    if a.ndim == 1 and b.ndim == 1:
        assert out is None
        if trans == 'c':
            return alpha * np.vdot(b, a)  # dotc conjugates *first* argument
        else:
            return alpha * a.dot(b)

    # Map all arrays to 2D arrays
    a = a.reshape(-1, a.shape[-1])
    if trans == 'n':
        b = b.reshape(b.shape[0], -1)
        outshape = a.shape[0], b.shape[1]
    else:  # 't' or 'c'
        b = b.reshape(-1, b.shape[-1])

    # Apply BLAS gemm routine
    outshape = a.shape[0], b.shape[trans == 'n']
    if out is None:
        # (ATLAS can't handle uninitialized output array)
        out = np.zeros(outshape, a.dtype)
    else:
        out = out.reshape(outshape)
    mmmx(alpha, a, 'N', b, trans.upper(), beta, out)

    # Determine actual shape of result array
    if trans == 'n':
        outshape = ashape[:-1] + bshape[1:]
    else:  # 't' or 'c'
        outshape = ashape[:-1] + bshape[:-1]
    return out.reshape(outshape)


if not hasattr(_gpaw, 'mmm'):
    def rk(alpha, a, beta, c, trans='c'):  # noqa
        if c.size == 0:
            return
        if beta == 0:
            c[:] = 0.0
        else:
            c *= beta
        if trans == 'n':
            c += alpha * a.conj().T.dot(a)
        else:
            a = a.reshape((len(a), -1))
            c += alpha * a.dot(a.conj().T)

    def r2k(alpha, a, b, beta, c, trans='c'):  # noqa
        if c.size == 0:
            return
        if beta == 0.0:
            c[:] = 0.0
        else:
            c *= beta
        if trans == 'c':
            c += (alpha * a.reshape((len(a), -1))
                  .dot(b.reshape((len(b), -1)).conj().T) +
                  alpha * b.reshape((len(b), -1))
                  .dot(a.reshape((len(a), -1)).conj().T))
        else:
            c += alpha * (a.conj().T @ b + b.conj().T @ a)

    def op(o, m):
        if o == 'N':
            return m
        if o == 'T':
            return m.T
        return m.conj().T

    def mmm(alpha: T, a: np.ndarray, opa: str,  # noqa
            b: np.ndarray, opb: str,
            beta: T, c: np.ndarray) -> None:
        if beta == 0.0:
            c[:] = 0.0
        else:
            c *= beta
        c += alpha * op(opa, a).dot(op(opb, b))

    gemmdot = _gemmdot

elif not debug:
    mmm = _gpaw.mmm  # noqa
    rk = _gpaw.rk  # noqa
    r2k = _gpaw.r2k  # noqa
    gemmdot = _gemmdot

else:
    def gemmdot(a, b, alpha=1.0, beta=1.0, out=None, trans='n'):
        assert a.flags.contiguous
        assert b.flags.contiguous
        assert a.dtype == b.dtype
        if trans == 'n':
            assert a.shape[-1] == b.shape[0]
        else:
            assert a.shape[-1] == b.shape[-1]
        if out is not None:
            assert out.flags.contiguous
            assert a.dtype == out.dtype
            assert a.ndim > 1 or b.ndim > 1
            if trans == 'n':
                assert out.shape == a.shape[:-1] + b.shape[1:]
            else:
                assert out.shape == a.shape[:-1] + b.shape[:-1]
        return _gemmdot(a, b, alpha, beta, out, trans)



import numpy as np
from ase.units import Ha

from gpaw.xc import XC
from gpaw.utilities import unpack


def vxc(gs, xc=None, coredensity=True, n1=0, n2=0):
    """Calculate XC-contribution to eigenvalues."""

    if n2 <= 0:
        n2 += gs.bd.nbands

    ham = gs.hamiltonian
    dens = gs.density

    if xc is None:
        xc = ham.xc
    elif isinstance(xc, str):
        xc = XC(xc)

    if dens.nt_sg is None:
        dens.interpolate_pseudo_density()

    thisisatest = not True

    if xc.orbital_dependent:
        gs.get_xc_difference(xc)

    # Calculate XC-potential:
    vxct_sg = ham.finegd.zeros(gs.nspins)
    xc.calculate(dens.finegd, dens.nt_sg, vxct_sg)
    vxct_sG = ham.restrict_and_collect(vxct_sg)
    if thisisatest:
        vxct_sG[:] = 1

    # ... and PAW corrections:
    dvxc_asii = {}
    for a, D_sp in dens.D_asp.items():
        dvxc_sp = np.zeros_like(D_sp)

        pawdata = gs.pawdatasets[a]
        if pawdata.hubbard_u is not None:
            _, dHU_sp = pawdata.hubbard_u.calculate(pawdata, D_sp)
            dvxc_sp += dHU_sp
        xc.calculate_paw_correction(pawdata, D_sp, dvxc_sp, a=a,
                                    addcoredensity=coredensity)

        dvxc_asii[a] = [unpack(dvxc_p) for dvxc_p in dvxc_sp]
        if thisisatest:
            dvxc_asii[a] = [pawdata.dO_ii]

    vxc_un = np.empty((gs.kd.mynk * gs.nspins, gs.bd.mynbands))
    for u, vxc_n in enumerate(vxc_un):
        kpt = gs.kpt_u[u]
        vxct_G = vxct_sG[kpt.s]
        for n in range(gs.bd.mynbands):
            if n1 <= n + gs.bd.beg < n2:
                psit_G = gs.get_wave_function_array(u, n)
                vxc_n[n] = gs.gd.integrate((psit_G * psit_G.conj()).real,
                                           vxct_G, global_integral=False)

        for a, dvxc_sii in dvxc_asii.items():
            m1 = max(n1, gs.bd.beg) - gs.bd.beg
            m2 = min(n2, gs.bd.end) - gs.bd.beg
            if m1 < m2:
                P_ni = kpt.P_ani[a][m1:m2]
                vxc_n[m1:m2] += ((P_ni @ dvxc_sii[kpt.s]) *
                                 P_ni.conj()).sum(1).real

    gs.gd.comm.sum(vxc_un)
    vxc_skn = gs.kd.collect(vxc_un, broadcast=True)

    if xc.orbital_dependent:
        vxc_skn += xc.exx_skn

    vxc_skn = gs.bd.collect(vxc_skn.T.copy(), broadcast=True).T
    return vxc_skn[:, :, n1:n2] * Ha



"""Calculate non self-consistent eigenvalues for hybrid functionals."""
from __future__ import annotations
import functools
import json
from pathlib import Path
from typing import Generator, List, Optional, Tuple, Union

import numpy as np
from ase.units import Ha
from gpaw.calculator import GPAW as GPAWOld
from gpaw import GPAW
from gpaw.new.ase_interface import ASECalculator
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.mpi import serial_comm
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.lfc import PWLFC
from gpaw.typing import Array3D
from gpaw.xc import XC
from gpaw.xc.kernel import XCNull
from gpaw.xc.tools import vxc

from . import parse_name
from .coulomb import coulomb_interaction
from .kpts import RSKPoint, get_kpt, to_real_space
from .paw import calculate_paw_stuff
from .symmetry import Symmetry


def non_self_consistent_eigenvalues(
        calc: Union[GPAWOld, ASECalculator, str, Path],
        xcname: str,
        n1: int = 0,
        n2: int = 0,
        kpt_indices: List[int] = None,
        snapshot: Union[str, Path] = None,
        ftol: float = 1e-9) -> tuple[Array3D,
                                     Array3D,
                                     Array3D]:
    """Calculate non self-consistent eigenvalues for a hybrid functional.

    Based on a self-consistent DFT calculation (calc).  Only eigenvalues n1 to
    n2 - 1 for the IBZ indices in kpt_indices are calculated
    (default is all bands and all k-points). EXX integrals involving
    states with occupation numbers less than ftol are skipped.  Use
    snapshot='name.json' to get snapshots for each k-point finished.

    Returns three (nspins, nkpts, n2 - n1)-shaped ndarrays
    with contributions to the eigenvalues in eV:

    >>> nsceigs = non_self_consistent_eigenvalues
    >>> eig_dft, vxc_dft, vxc_hyb = nsceigs('<gpw-file>', xcname='PBE0')
    >>> eig_hyb = eig_dft - vxc_dft + vxc_hyb
    """

    if not isinstance(calc, (GPAWOld, ASECalculator)):
        if calc == '<gpw-file>':  # for doctest
            return (np.zeros((1, 1, 1)),
                    np.zeros((1, 1, 1)),
                    np.zeros((1, 1, 1)))
        calc = GPAW(Path(calc), txt=None, parallel={'band': 1, 'kpt': 1})

    assert isinstance(calc, (GPAWOld, ASECalculator))
    wfs = calc.wfs

    if n2 <= 0:
        n2 += wfs.bd.nbands

    if kpt_indices is None:
        kpt_indices = np.arange(wfs.kd.nibzkpts).tolist()

    path = Path(snapshot) if snapshot is not None else None

    e_dft_sin = np.zeros(0)
    v_dft_sin = np.zeros(0)

    # sl=semilocal, nl=nonlocal
    v_hyb_sl_sin = np.zeros(0)
    v_hyb_nl_sin: Optional[List[List[np.ndarray]]] = None

    if path:
        e_dft_sin, v_dft_sin, v_hyb_sl_sin, v_hyb_nl_sin = read_snapshot(path)

    xcname, exx_fraction, omega = parse_name(xcname)

    if v_dft_sin.size == 0:
        xc = XC(xcname)
        e_dft_sin, v_dft_sin, v_hyb_sl_sin = _semi_local(
            calc, xc, n1, n2, kpt_indices)
        write_snapshot(e_dft_sin, v_dft_sin, v_hyb_sl_sin, v_hyb_nl_sin,
                       path, wfs.world)
    # Non-local hybrid contribution
    if v_hyb_nl_sin is None:
        v_hyb_nl_sin = [[] for s in range(wfs.nspins)]

    # Find missing indices:
    kpt_indices_s = [kpt_indices[len(v_hyb_nl_in):]
                     for v_hyb_nl_in in v_hyb_nl_sin]

    if any(len(kpt_indices) > 0 for kpt_indices in kpt_indices_s):
        for s, v_hyb_nl_n in _non_local(calc, n1, n2, kpt_indices_s,
                                        ftol, omega):
            v_hyb_nl_sin[s].append(v_hyb_nl_n * exx_fraction)
            write_snapshot(e_dft_sin, v_dft_sin, v_hyb_sl_sin, v_hyb_nl_sin,
                           path, wfs.world)

    return (e_dft_sin * Ha,
            v_dft_sin * Ha,
            (v_hyb_sl_sin + v_hyb_nl_sin) * Ha)


def _semi_local(calc: GPAWOld | ASECalculator,
                xc,
                n1: int,
                n2: int,
                kpt_indices: List[int]
                ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    wfs = calc.wfs
    nspins = wfs.nspins
    e_dft_sin = np.array([[calc.get_eigenvalues(k, spin)[n1:n2]
                           for k in kpt_indices]
                          for spin in range(nspins)])
    v_dft_sin = vxc(calc.gs_adapter(), n1=n1, n2=n2)[:, kpt_indices]
    if isinstance(xc.kernel, XCNull):
        v_hyb_sl_sin = np.zeros_like(v_dft_sin)
    else:
        v_hyb_sl_sin = vxc(calc.gs_adapter(), xc, n1=n1, n2=n2)[:, kpt_indices]
    return e_dft_sin / Ha, v_dft_sin / Ha, v_hyb_sl_sin / Ha


def _non_local(calc: GPAWOld | ASECalculator,
               n1: int,
               n2: int,
               kpt_indices_s: List[List[int]],
               ftol: float,
               omega: float) -> Generator[Tuple[int, np.ndarray], None, None]:
    wfs = calc.wfs
    kd = wfs.kd
    dens = calc.density

    nocc = max(((kpt.f_n / kpt.weight) > ftol).sum()
               for kpt in wfs.kpt_u)
    nocc = kd.comm.max(wfs.bd.comm.sum(int(nocc)))

    coulomb = coulomb_interaction(omega, wfs.gd, kd)
    sym = Symmetry(kd)

    paw_s = calculate_paw_stuff(wfs, dens)

    for spin, kpt_indices in enumerate(kpt_indices_s):
        if len(kpt_indices) == 0:
            continue
        kpts2 = [get_kpt(wfs, k, spin, 0, nocc) for k in range(kd.nibzkpts)]
        for i in kpt_indices:
            kpt1 = get_kpt(wfs, i, spin, n1, n2)
            v_n = _calculate_eigenvalues(
                kpt1, kpts2, paw_s[spin], kd, coulomb, sym, wfs, calc.spos_ac)
            wfs.world.sum(v_n)
            yield spin, v_n


def _calculate_eigenvalues(kpt1, kpts2, paw, kd, coulomb, sym, wfs, spos_ac):
    pd = kpt1.psit.pd
    gd = pd.gd.new_descriptor(comm=serial_comm)
    comm = wfs.world
    size = comm.size
    rank = comm.rank

    nsym = len(kd.symmetry.op_scc)
    assert len(kpts2) == kd.nibzkpts

    N1 = len(kpt1.psit.array)
    N2 = len(kpts2[0].psit.array)

    size1, size2 = layout(N1, N2, size)
    assert size1 * size2 == size
    B1 = (N1 + size1 - 1) // size1
    B2 = (N2 + size2 - 1) // size2
    rank1, rank2 = divmod(rank, size2)
    n1a = min(B1 * rank1, N1)
    n1b = min(n1a + B1, N1)
    n2a = min(B2 * rank2, N2)
    n2b = min(n2a + B2, N2)

    u1_nR = to_real_space(kpt1.psit, n1a, n1b)
    proj1all = kpt1.proj.broadcast()
    proj1 = proj1all.view(n1a, n1b)

    E_n = np.zeros(N1)
    e_n = E_n[n1a:n1b]
    e_nn = np.empty((n1b - n1a, n2b - n2a))

    for i2, kpt2 in enumerate(kpts2):
        u2_nR = to_real_space(kpt2.psit, n2a, n2b)
        rskpt0 = RSKPoint(u2_nR,
                          kpt2.proj.broadcast().view(n2a, n2b),
                          kpt2.f_n[n2a:n2b],
                          kpt2.k_c,
                          kpt2.weight)
        for k, i in enumerate(kd.bz2ibz_k):
            if i != i2:
                continue
            s = kd.sym_k[k] + kd.time_reversal_k[k] * nsym
            rskpt2 = sym.apply_symmetry(s, rskpt0, wfs, spos_ac)
            q_c = rskpt2.k_c - kpt1.k_c
            qd = KPointDescriptor([-q_c])
            pd12 = PWDescriptor(pd.ecut, gd, pd.dtype, kd=qd)
            ghat = PWLFC([data.ghat_l for data in wfs.setups], pd12)
            ghat.set_positions(spos_ac)
            v_G = coulomb.get_potential(pd12)
            Q_annL = [np.einsum('mi, ijL, nj -> mnL',
                                proj1[a],
                                Delta_iiL,
                                rskpt2.proj[a].conj())
                      for a, Delta_iiL in enumerate(paw.Delta_aiiL)]
            rho_nG = ghat.pd.empty(n2b - n2a, u1_nR.dtype)

            for n1, u1_R in enumerate(u1_nR):
                for u2_R, rho_G in zip(rskpt2.u_nR, rho_nG):
                    rho_G[:] = ghat.pd.fft(u1_R * u2_R.conj())

                ghat.add(rho_nG,
                         {a: Q_nnL[n1] for a, Q_nnL in enumerate(Q_annL)})

                for n2, rho_G in enumerate(rho_nG):
                    vrho_G = v_G * rho_G
                    e = ghat.pd.integrate(rho_G, vrho_G).real
                    e_nn[n1, n2] = e / kd.nbzkpts
            e_n -= e_nn.dot(rskpt2.f_n)

    for a, VV_ii in paw.VV_aii.items():
        P_ni = proj1all[a]
        vv_n = np.einsum('ni, ij, nj -> n',
                         P_ni.conj(), VV_ii, P_ni).real
        vc_n = np.einsum('ni, ij, nj -> n',
                         P_ni.conj(), paw.VC_aii[a], P_ni).real
        E_n -= (2 * vv_n + vc_n)

    return E_n


def write_snapshot(e_dft_sin: np.ndarray,
                   v_dft_sin: np.ndarray,
                   v_hyb_sl_sin: np.ndarray,
                   v_hyb_nl_sin: Optional[List[List[np.ndarray]]],
                   path: Optional[Path],
                   comm) -> None:
    """Write to json-file what has been calculated so far."""
    if comm.rank == 0 and path:
        dct = {'e_dft_sin': e_dft_sin.tolist(),
               'v_dft_sin': v_dft_sin.tolist(),
               'v_hyb_sl_sin': v_hyb_sl_sin.tolist()}
        if v_hyb_nl_sin is not None:
            dct['v_hyb_nl_sin'] = [[v_n.tolist()
                                    for v_n in v_in]
                                   for v_in in v_hyb_nl_sin]
        path.write_text(json.dumps(dct, indent=0))


def read_snapshot(snapshot: Path
                  ) -> Tuple[np.ndarray,
                             np.ndarray,
                             np.ndarray,
                             Optional[List[List[np.ndarray]]]]:
    """Read from json-file what has already been calculated."""
    if snapshot.is_file():
        dct = json.loads(snapshot.read_text())
        v_hyb_nl_sin = dct.get('v_hyb_nl_sin')
        if v_hyb_nl_sin is not None:
            v_hyb_nl_sin = [[np.array(v_n)
                             for v_n in v_in]
                            for v_in in v_hyb_nl_sin]
        return (np.array(dct['e_dft_sin']),
                np.array(dct['v_dft_sin']),
                np.array(dct['v_hyb_sl_sin']),
                v_hyb_nl_sin)
    return np.array([[[]]]), np.array([[[]]]), np.array([[[]]]), None


@functools.lru_cache()
def layout(n1: int, n2: int, size: int) -> Tuple[int, int]:
    """Distribute n1*n2 matrix over s1*s2=size blocks.

    Returns s1, s2.

    >>> layout(10, 10, 8)
    (4, 2)
    """
    candidates: List[Tuple[float, int, int]] = []
    for s1 in range(1, size + 1):
        s2, r = divmod(size, s1)
        if r > 0:
            continue
        fitness = (1 - idle(n1, s1)) * (1 - idle(n2, s2))
        candidates.append((fitness, s1, s2))
    return max(candidates)[1:]


def idle(n: int, s: int) -> float:
    """Idle fraction (helper function for layout() function)."""
    b = (n + s - 1) // s
    return 1 - n / (b * s)



from math import pi

import numpy as np

from gpaw.hybrids.wstc import WignerSeitzTruncatedCoulomb as WSTC


def coulomb_interaction(omega, gd, kd):
    if omega:
        return ShortRangeCoulomb(omega)
    return WSTC(gd.cell_cv, kd.N_c)


class ShortRangeCoulomb:
    def __init__(self, omega):
        self.omega = omega

    def get_description(self):
        return f'Short-range Coulomb: erfc(omega*r)/r (omega = {self.omega} ' \
               'bohr^-1)'
    
    def get_potential(self, pd):
        G2_G = pd.G2_qG[0]
        x_G = 1 - np.exp(-G2_G / (4 * self.omega**2))
        with np.errstate(invalid='ignore'):
            v_G = 4 * pi * x_G / G2_G
        G0 = G2_G.argmin()
        if G2_G[G0] < 1e-11:
            v_G[G0] = pi / self.omega**2
        return v_G



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from gpaw.new.calculation import DFTState


class Eigensolver:
    def iterate(self, state: DFTState, hamiltonian) -> float:
        raise NotImplementedError



import pytest
import numpy as np
from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.response.bse import BSE, read_bse_eigenvalues
from gpaw.response.df import read_response_function
from gpaw.test import findpeak, equal


@pytest.mark.response
def test_response_bse_silicon(in_tmp_dir, scalapack):
    GS = 1
    nosym = 1
    bse = 1
    check = 1

    if GS:
        a = 5.431  # From PRB 73,045112 (2006)
        atoms = bulk('Si', 'diamond', a=a)
        atoms.positions -= a / 8
        calc = GPAW(mode='pw',
                    kpts={'size': (2, 2, 2), 'gamma': True},
                    occupations=FermiDirac(0.001),
                    nbands=12,
                    convergence={'bands': -4})
        atoms.calc = calc
        atoms.get_potential_energy()
        calc.write('Si.gpw', 'all')

    if bse:
        eshift = 0.8
        bse = BSE('Si.gpw',
                  ecut=50.,
                  valence_bands=range(4),
                  conduction_bands=range(4, 8),
                  eshift=eshift,
                  nbands=8,
                  write_h=False,
                  write_v=False)
        bse.get_dielectric_function(eta=0.2,
                                    w_w=np.linspace(0, 10, 2001))
        w_w, epsreal_w, epsimag_w = read_response_function('df_bse.csv')
    if check:
        w_ = 2.552
        I_ = 421.15
        w, I = findpeak(w_w, epsimag_w)
        equal(w, w_, 0.01)
        equal(I, I_, 0.1)

    if GS and nosym:
        atoms = bulk('Si', 'diamond', a=a)
        calc = GPAW(mode='pw',
                    kpts={'size': (2, 2, 2), 'gamma': True},
                    occupations=FermiDirac(0.001),
                    nbands=12,
                    symmetry='off',
                    convergence={'bands': -4})
        atoms.calc = calc
        atoms.get_potential_energy()
        calc.write('Si.gpw', 'all')

    if bse and nosym:
        bse = BSE('Si.gpw',
                  ecut=50.,
                  valence_bands=range(4),
                  conduction_bands=range(4, 8),
                  eshift=eshift,
                  nbands=8,
                  write_h=False,
                  write_v=False)
        w_w, eps_w = bse.get_dielectric_function(filename=None,
                                                 eta=0.2,
                                                 w_w=np.linspace(0, 10, 2001))

    if check and nosym:
        w, I = findpeak(w_w, eps_w.imag)
        equal(w, w_, 0.01)
        equal(I, I_, 0.1)

        # Read eigenvalues file and test first 3 weights:
        _, C_w = read_bse_eigenvalues('eig.dat')
        assert C_w[0] == pytest.approx(22.37, abs=0.05)
        # These two have degenerate eigenvalues:
        assert np.sum(C_w[1:3]) == pytest.approx(44.29, abs=0.05)



"""

tests solvation parameters as in

V. M. Sanchez, M. Sued, and D. A. Scherlis,
The Journal of Chemical Physics, vol. 131, no. 17, p. 174108, 2009
"""

from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.test import equal
from ase import Atoms
from ase.units import mol, kcal, Pascal, m, Bohr
from gpaw.solvation import (
    SolvationGPAW,
    FG02SmoothStepCavity,
    LinearDielectric,
    GradientSurface,
    SurfaceInteraction,
    SSS09Density)


def test_solvation_sss09():
    SKIP_VAC_CALC = True

    h = 0.24
    vac = 4.0

    epsinf = 78.36
    rho0 = 1.0 / Bohr ** 3
    beta = 2.4
    st = 72. * 1e-3 * Pascal * m

    def atomic_radii(atoms):
        return [2.059]

    convergence = {
        'energy': 0.05 / 8.,
        'density': 10.,
        'eigenstates': 10.,
    }

    atoms = Cluster(Atoms('Cl'))
    atoms.minimal_box(vac, h)

    if not SKIP_VAC_CALC:
        atoms.calc = GPAW(xc='PBE', h=h, charge=-1, convergence=convergence)
        Evac = atoms.get_potential_energy()
        print(Evac)
    else:
        # h=0.24, vac=4.0, setups: 0.9.11271, convergence: only energy 0.05 / 8
        Evac = -3.83245253419

    atoms.calc = SolvationGPAW(
        xc='PBE', h=h, charge=-1, convergence=convergence,
        cavity=FG02SmoothStepCavity(
            rho0=rho0, beta=beta,
            density=SSS09Density(atomic_radii=atomic_radii),
            surface_calculator=GradientSurface()
        ),
        dielectric=LinearDielectric(epsinf=epsinf),
        interactions=[SurfaceInteraction(surface_tension=st)]
    )
    Ewater = atoms.get_potential_energy()
    assert atoms.calc.get_number_of_iterations() < 40
    atoms.get_forces()
    DGSol = (Ewater - Evac) / (kcal / mol)
    print('Delta Gsol: %s kcal / mol' % DGSol)

    equal(DGSol, -75., 10.)



import numpy as np

from ase.utils import IOContext

from gpaw.lcaotddft.observer import TDDFTObserver
from gpaw.utilities.scalapack import scalapack_zero


class EnergyWriter(TDDFTObserver):
    version = 1

    def __init__(self, paw, dmat, filename, interval=1):
        TDDFTObserver.__init__(self, paw, interval)
        self.ioctx = IOContext()
        self.dmat = dmat
        if paw.niter == 0:
            # Initialize
            self.fd = self.ioctx.openfile(filename, comm=paw.world, mode='w')
        else:
            # Read and continue
            self.fd = self.ioctx.openfile(filename, comm=paw.world, mode='a')

    def _write(self, line):
        self.fd.write(line)
        self.fd.flush()

    def _write_header(self, paw):
        if paw.niter != 0:
            return
        line = '# %s[version=%s]\n' % (self.__class__.__name__, self.version)
        line += ('# %15s %22s %22s %22s %22s %22s %22s\n' %
                 ('time', 'kinetic0', 'coulomb', 'zero', 'external',
                  'xc', 'band'))
        self._write(line)

    def _write_kick(self, paw):
        time = paw.time
        kick = paw.kick_strength
        line = '# Kick = [%22.12le, %22.12le, %22.12le]; ' % tuple(kick)
        line += 'Time = %.8lf\n' % time
        self._write(line)

    def _get_energies(self, paw):
        e_band = 0.0
        rho_uMM = self.dmat.get_density_matrix((paw.niter, paw.action))
        get_H_MM = paw.td_hamiltonian.get_hamiltonian_matrix
        ksl = paw.wfs.ksl
        for u, kpt in enumerate(paw.wfs.kpt_u):
            rho_MM = rho_uMM[u]

            # H_MM = get_H_MM(kpt, paw.time)
            H_MM = get_H_MM(kpt, paw.time, addfxc=False, addpot=False)

            if ksl.using_blacs:
                # rhoH_MM = (rho_MM * H_MM).real  # General case
                rhoH_MM = rho_MM.real * H_MM.real  # Hamiltonian is real
                # Hamiltonian has correct values only in lower half, so
                # 1. Add lower half and diagonal twice
                scalapack_zero(ksl.mmdescriptor, rhoH_MM, 'U')
                e = 2 * np.sum(rhoH_MM)
                # 2. Reduce the extra diagonal
                scalapack_zero(ksl.mmdescriptor, rhoH_MM, 'L')
                e -= np.sum(rhoH_MM)
                # Sum over all ranks
                e = ksl.block_comm.sum(e)
            else:
                e = np.sum(rho_MM * H_MM).real

            e_band += e

        paw.occupations.e_band = e_band
        paw.occupations.e_entropy = 0.0
        e_kinetic0 = paw.hamiltonian.e_kinetic0
        e_coulomb = paw.hamiltonian.e_coulomb
        e_zero = paw.hamiltonian.e_zero
        e_external = paw.hamiltonian.e_external
        e_xc = paw.hamiltonian.e_xc

        return np.array((e_kinetic0, e_coulomb, e_zero,
                         e_external, e_xc, e_band))

    def _write_energy(self, paw):
        time = paw.time
        energy_i = self._get_energies(paw) - self.energy0_i
        line = (
            '%20.8lf %22.12le %22.12le %22.12le %22.12le %22.12le %22.12le\n' %
            ((time, ) + tuple(energy_i)))
        self._write(line)

    def _update(self, paw):
        if paw.action == 'init':
            self._write_header(paw)
            self.energy0_i = self._get_energies(paw)
        elif paw.action == 'kick':
            self._write_kick(paw)
        self._write_energy(paw)

    def __del__(self):
        self.ioctx.close()



"""Compute site-kernels. Used for computing Heisenberg exchange constants.
Specifically, one maps a DFT calculations onto a Heisenberg lattice model,
where the site kernels define the lattice sites and magnetic moments."""

import numpy as np
from scipy.special import jv
from gpaw.response.pair_functions import get_pw_coordinates
from ase.units import Bohr


class SiteKernels:
    """Factory for calculating sublattice site kernels

                1  /
    K_aGG'(q) = ‾‾ | dr e^(-i[G-G'+q].r) θ(r∊V_a)
                V0 /

    where V_a denotes the integration volume of site a, centered at the site
    position τ_a, and V0 denotes the unit cell volume."""

    def __init__(self, positions, partitions):
        """Construct the site kernel factory.

        Parameters
        ----------
        positions : np.ndarray
            Site positions in a.u. (Bohr). Shape: (nsites, 3).
        partitions : list
            List (len=npartitions) of lists (len=nsites) with site geometries
            where the geometry arguments are given in atomic units.
            For more information on site geometries, see calculate_site_kernels
        """
        assert isinstance(partitions, list)
        assert all([isinstance(geometries, list)
                    and len(geometries) == positions.shape[0]
                    for geometries in partitions])
        assert positions.shape[1] == 3

        self.positions = positions
        self.partitions = partitions

    @property
    def npartitions(self):
        return len(self.partitions)

    @property
    def nsites(self):
        return self.positions.shape[0]

    @property
    def shape(self):
        return self.npartitions, self.nsites

    @property
    def geometry_shapes(self):
        """If all sites of a given partition has the same geometry, the
        partition is said to have a geometry shape. Otherwise, the
        geometry shape is None."""

        geometry_shapes = []
        for geometries in self.partitions:
            # Record the geometry shape, if they are all the same
            if all([shape == geometries[0][0] for shape, _ in geometries]):
                geometry_shapes.append(geometries[0][0])
            else:
                geometry_shapes.append(None)

        return geometry_shapes

    def calculate(self, qpd):
        """Generate the site kernels of each partition.

        Returns
        -------
        K_aGG : np.ndarray (dtype=complex)
            Site kernels of sites a and plane wave components G and G'.
        """
        for geometries in self.partitions:
            # We yield one set of site kernels at a time, because they can be
            # memory intensive
            yield calculate_site_kernels(qpd, self.positions, geometries)

    def __add__(self, sitekernels):
        """Add the sites from two SiteKernels instances to a new joint
        SiteKernels instance with nsites = nsites1 + nsites2."""
        assert isinstance(sitekernels, SiteKernels)
        assert self.npartitions == sitekernels.npartitions

        # Join positions
        nsites = self.nsites + sitekernels.nsites
        positions = np.append(self.positions,
                              sitekernels.positions).reshape(nsites, 3)

        # Join partitions
        partitions = [geometries1 + geometries2
                      for geometries1, geometries2
                      in zip(self.partitions, sitekernels.partitions)]

        return SiteKernels(positions, partitions)

    def append(self, sitekernels):
        """Append the partitions of another array with identical site
        positions such that npartitions = npartitions1 + npartitions2."""
        assert isinstance(sitekernels, SiteKernels)
        assert self.nsites == sitekernels.nsites
        assert np.allclose(self.positions, sitekernels.positions)

        self.partitions += sitekernels.partitions

    def copy(self):
        return SiteKernels(self.positions.copy(), self.partitions.copy())


class SphericalSiteKernels(SiteKernels):

    def __init__(self, positions, radii):
        """Construct a site kernel factory with spherical kernels.

        Parameters
        ----------
        positions : np.ndarray
            Site positions in Angstrom (Å). Shape: (nsites, 3).
        radii : list or np.ndarray
            Spherical radii of the sites in Angstrom (Å).
            Shape: (npartitions, nsites), where the individual spherical radii
            can be varried for each spatial partitioning.
        """
        positions = np.asarray(positions)

        # Parse the input spherical radii
        rc_pa = np.asarray(radii)
        assert len(rc_pa.shape) == 2
        assert rc_pa.shape[1] == positions.shape[0]

        # Convert radii to internal units (Å to Bohr)
        positions = positions / Bohr
        rc_pa = rc_pa / Bohr

        # Generate partitions as list of lists of geometries
        partitions = [[('sphere', (rc,)) for rc in rc_a]
                      for rc_a in rc_pa]

        SiteKernels.__init__(self, positions, partitions)


class CylindricalSiteKernels(SiteKernels):

    def __init__(self, positions, directions, radii, heights):
        """Construct a site kernel factory with cylindrical kernels.

        Parameters
        ----------
        positions : np.ndarray
            Site positions in Angstrom (Å). Shape: (nsites, 3).
        directions : np.ndarray
            Normalized directions of the cylindrical axes.
            Shape: (npartitions, nsites, 3), where the direction of each
            individual cylinder can be varried (along with the radius and
            height) for each spatial partitioning.
        radii : np.ndarray
            Cylindrical radii of the sites in Angstrom (Å).
            Shape: (npartitions, nsites).
        heights : list or np.ndarray
            Cylinder heights in Angstrom (Å). Shape: (npartitions, nsites).
        """
        positions = np.asarray(positions)

        # Parse the cylinder geometry arguments
        ez_pav = np.asarray(directions)
        rc_pa = np.asarray(radii)
        hc_pa = np.asarray(heights)
        nsites = positions.shape[0]
        npartitions = ez_pav.shape[0]
        assert ez_pav.shape == (npartitions, nsites, 3)
        assert np.allclose(np.linalg.norm(ez_pav, axis=-1), 1., atol=1.e-8)
        assert rc_pa.shape == (npartitions, nsites)
        assert hc_pa.shape == (npartitions, nsites)

        # Convert to internal units (Å to Bohr)
        positions = positions / Bohr
        rc_pa = rc_pa / Bohr
        hc_pa = hc_pa / Bohr

        # Generate partitions as list of lists of geometries
        partitions = [[('cylinder', (ez_v, rc, hc))
                       for ez_v, rc, hc in zip(ez_av, rc_a, hc_a)]
                      for ez_av, rc_a, hc_a in zip(ez_pav, rc_pa, hc_pa)]

        SiteKernels.__init__(self, positions, partitions)


class ParallelepipedicSiteKernels(SiteKernels):

    def __init__(self, positions, cells):
        """Construct a site kernel factory with parallelepipedic kernels.

        Parameters
        ----------
        positions : np.ndarray
            Site positions in Angstrom (Å). Shape: (nsites, 3).
        cells : np.ndarray
            Cell vectors of the parallelepiped in Angstrom (Å).
            Shape: (npartitions, nsites, 3, 3), where the parallelepipedic
            cell of each site can be varried independently for each spatial
            partitioning. The second to last entry is the vector index and
            the last entry indexes the cartesian components.
        """
        positions = np.asarray(positions)

        # Parse the parallelepipeds' cells
        cell_pacv = np.asarray(cells)
        assert len(cell_pacv.shape) == 4
        assert cell_pacv.shape[1:] == (positions.shape[0], 3, 3)

        # Convert to internal units (Å to Bohr)
        positions = positions / Bohr
        cell_pacv = cell_pacv / Bohr

        # Generate partitions as list of lists of geometries
        partitions = [[('parallelepiped', (cell_cv,)) for cell_cv in cell_acv]
                      for cell_acv in cell_pacv]

        SiteKernels.__init__(self, positions, partitions)


def calculate_site_kernels(qpd, positions, geometries):
    """Calculate the sublattice site kernel:

                1  /
    K_aGG'(q) = ‾‾ | dr e^(-i[G-G'+q].r) θ(r∊V_a)
                V0 /

    where V_a denotes the integration volume of site a, centered at the site
    position τ_a, and V0 denotes the unit cell volume.

    In the calculation, the kernel is split in two contributions:

    1) The Fourier component of the site position:

    τ_a(Q) = e^(-iQ.τ_a)

    2) The site centered geometry factor

           /
    Θ(Q) = | dr e^(-iQ.r) θ(r+τ_a∊V_a)
           /

    where Θ(Q) only depends on the geometry of the integration volume.
    With this:

                1
    K_aGG'(q) = ‾‾ τ_a(G-G'+q) Θ(G-G'+q)
                V0

    Parameters
    ----------
    qpd : SingleQPWDescriptor
        Plane wave descriptor corresponding to the q wave vector of interest.
    positions : np.ndarray
        Site positions. Array of shape (nsites, 3).
    geometries : list
        List of site geometries. A site geometry is a tuple of the integration
        volume shape (str) and arguments (tuple): (shape, args). Valid shapes
        are 'sphere', 'cylinder' and 'parallelepiped'. The integration volume
        arguments specify the size and orientation of the integration region.

    Returns
    -------
    K_aGG : np.ndarray (dtype=complex)
        Site kernels of sites a and plane wave components G and G'.
    """
    assert positions.shape[0] == len(geometries)
    assert positions.shape[1] == 3

    # Extract unit cell volume
    V0 = qpd.gd.volume

    # Construct Q=G-G'+q
    Q_GGv = construct_wave_vectors(qpd)

    # Allocate site kernel array
    nsites = len(geometries)
    K_aGG = np.zeros((nsites,) + Q_GGv.shape[:2], dtype=complex)

    # Calculate the site kernel for each site individually
    for a, (tau_v, (shape, args)) in enumerate(zip(positions, geometries)):

        # Compute the site centered geometry factor
        _geometry_factor = create_geometry_factor(shape)  # factory pattern
        Theta_GG = _geometry_factor(Q_GGv, *args)

        # Compute the Fourier component of the site position
        tau_GG = np.exp(-1.j * Q_GGv @ tau_v)

        # Update data
        K_aGG[a, :, :] = 1 / V0 * tau_GG * Theta_GG

    return K_aGG


def construct_wave_vectors(qpd):
    """Construct wave vectors Q=G1-G2+q corresponding to the q-vector of
    interest."""
    G_Gv, q_v = get_plane_waves_and_reduced_wave_vector(qpd)

    # Allocate arrays for G, G' and q respectively
    nG = len(G_Gv)
    G1_GGv = np.tile(G_Gv[:, np.newaxis, :], [1, nG, 1])
    G2_GGv = np.tile(G_Gv[np.newaxis, :, :], [nG, 1, 1])
    q_GGv = np.tile(q_v[np.newaxis, np.newaxis, :], [nG, nG, 1])

    # Contruct the wave vector G1 - G2 + q
    Q_GGv = G1_GGv - G2_GGv + q_GGv

    return Q_GGv


def get_plane_waves_and_reduced_wave_vector(qpd):
    """Get the reciprocal lattice vectors and reduced wave vector of the plane
    wave representation corresponding to the q-vector of interest."""
    # Get the reduced wave vector
    q_c = qpd.q_c

    # Get the reciprocal lattice vectors in relative coordinates
    G_Gc = get_pw_coordinates(qpd)

    # Convert to cartesian coordinates
    B_cv = 2.0 * np.pi * qpd.gd.icell_cv  # Coordinate transform matrix
    q_v = np.dot(q_c, B_cv)  # Unit = Bohr^(-1)
    G_Gv = np.dot(G_Gc, B_cv)

    return G_Gv, q_v


def create_geometry_factor(shape):
    """Creator component of the geometry factor factory pattern."""
    if shape == 'sphere':
        return spherical_geometry_factor
    elif shape == 'cylinder':
        return cylindrical_geometry_factor
    elif shape == 'parallelepiped':
        return parallelepipedic_geometry_factor

    raise ValueError('Invalid site kernel shape:', shape)


def spherical_geometry_factor(Q_Qv, rc):
    """Calculate the site centered geometry factor for a spherical site kernel:

           /
    Θ(Q) = | dr e^(-iQ.r) θ(|r|<r_c)
           /

           4πr_c
         = ‾‾‾‾‾ [sinc(|Q|r_c) - cos(|Q|r_c)]
           |Q|^2

                    3 [sinc(|Q|r_c) - cos(|Q|r_c)]
         = V_sphere ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                              (|Q|r_c)^2

    where the dimensionless geometry factor satisfies:

    Θ(Q)/V_sphere --> 1 for |Q|r_c --> 0.

    Parameters
    ----------
    Q_Qv : np.ndarray
        Wave vectors to evaluate the site centered geometry factor at. The
        cartesian coordinates needs to be the last dimension of the array (v),
        but the preceeding index/indices Q can have any tensor structure, such
        that Q_Qv.shape = (..., 3).
    rc : float
        Radius of the sphere.
    """
    assert Q_Qv.shape[-1] == 3
    assert isinstance(rc, float) and rc > 0.

    # Calculate the sphere volume
    Vsphere = 4 * np.pi * rc**3. / 3

    # Calculate |Q|r_c
    Qrc_Q = np.linalg.norm(Q_Qv, axis=-1) * rc

    # Allocate array with ones to provide the correct dimensionless geometry
    # factor in the |Q|r_c --> 0 limit.
    # This is done to avoid division by zero.
    Theta_Q = np.ones(Q_Qv.shape[:-1], dtype=float)

    # Calculate the dimensionless geometry factor
    Qrcs = Qrc_Q[Qrc_Q > 1.e-8]
    Theta_Q[Qrc_Q > 1.e-8] = 3. * (sinc(Qrcs) - np.cos(Qrcs)) / Qrcs**2.

    Theta_Q *= Vsphere

    return Theta_Q


def cylindrical_geometry_factor(Q_Qv, ez_v, rc, hc):
    """Calculate site centered geometry factor for a cylindrical site kernel:

           /
    Θ(Q) = | dr e^(-iQ.r) θ(ρ<r_c) θ(|z|/2<h_c)
           /

            4πr_c
         = ‾‾‾‾‾‾‾ J_1(Q_ρ r_c) sin(Q_z h_c / 2)
           Q_ρ Q_z

                      2 J_1(Q_ρ r_c)
         = V_cylinder ‾‾‾‾‾‾‾‾‾‾‾‾‾‾ sinc(Q_z h_c / 2)
                         Q_ρ r_c

    where z denotes the cylindrical axis, ρ the radial axis and the
    dimensionless geometry factor satisfy:

    Θ(Q)/V_cylinder --> 1 for Q --> 0.

    Parameters
    ----------
    Q_Qv : np.ndarray
        Wave vectors to evaluate the site centered geometry factor at. The
        cartesian coordinates needs to be the last dimension of the array (v),
        but the preceeding index/indices Q can have any tensor structure, such
        that Q_Qv.shape = (..., 3).
    ez_v : np.ndarray
        Normalized direction of the cylindrical axis.
    rc : float
        Radius of the cylinder.
    hc : float
        Height of the cylinder.
    """
    assert Q_Qv.shape[-1] == 3
    assert ez_v.shape == (3,)
    assert abs(np.linalg.norm(ez_v) - 1.) < 1.e-8
    assert isinstance(rc, float) and rc > 0.
    assert isinstance(hc, float) and hc > 0.

    # Calculate cylinder volume
    Vcylinder = np.pi * rc**2. * hc

    # Calculate Q_z h_c and Q_ρ r_c
    Qzhchalf_Q = np.abs(Q_Qv @ ez_v) * hc / 2.
    Qrhorc_Q = np.linalg.norm(np.cross(Q_Qv, ez_v), axis=-1) * rc

    # Allocate array with ones to provide the correct dimensionless geometry
    # factor in the Q_ρ r_c --> 0 limit.
    # This is done to avoid division by zero.
    Theta_Q = np.ones(Q_Qv.shape[:-1], dtype=float)

    # Calculate the dimensionless geometry factor
    Qrhorcs = Qrhorc_Q[Qrhorc_Q > 1.e-8]
    Theta_Q[Qrhorc_Q > 1.e-8] = 2. * jv(1, Qrhorcs) / Qrhorcs
    Theta_Q *= Vcylinder * sinc(Qzhchalf_Q)

    return Theta_Q


def parallelepipedic_geometry_factor(Q_Qv, cell_cv):
    """Calculate the site centered geometry factor for a parallelepipedic site
    kernel:

           /
    Θ(Q) = | dr e^(-iQ.r) θ(r∊V_parallelepiped)
           /

         = |det[a1, a2, a3]| sinc(Q.a1 / 2) sinc(Q.a2 / 2) sinc(Q.a3 / 2)

         = V_parallelepiped sinc(Q.a1 / 2) sinc(Q.a2 / 2) sinc(Q.a3 / 2)

    where a1, a2 and a3 denotes the parallelepipedic cell vectors.

    Parameters
    ----------
    Q_Qv : np.ndarray
        Wave vectors to evaluate the site centered geometry factor at. The
        cartesian coordinates needs to be the last dimension of the array (v),
        but the preceeding index/indices Q can have any tensor structure, such
        that Q_Qv.shape = (..., 3).
    cell_cv : np.ndarray, shape=(3, 3)
        Cell vectors of the parallelepiped, where v denotes the cartesian
        coordinates.
    """
    assert Q_Qv.shape[-1] == 3
    assert cell_cv.shape == (3, 3)

    # Calculate the parallelepiped volume
    Vparlp = abs(np.linalg.det(cell_cv))
    assert Vparlp > 1.e-8  # Not a valid parallelepiped if volume vanishes

    # Calculate the site-kernel
    a1, a2, a3 = cell_cv
    Theta_Q = Vparlp * sinc(Q_Qv @ a1 / 2) * sinc(Q_Qv @ a2 / 2) * \
        sinc(Q_Qv @ a3 / 2)

    return Theta_Q


def sinc(x):
    """np.sinc(x) = sin(pi*x) / (pi*x), hence the division by pi"""
    return np.sinc(x / np.pi)



from time import localtime
import pickle

import numpy as np
from ase.units import Ha
from ase.calculators.singlepoint import SinglePointCalculator
from ase.utils import IOContext

from gpaw.utilities import pack
from gpaw.utilities.tools import tri2full
from gpaw.utilities.blas import rk, mmm, mmmx
from gpaw.basis_data import Basis
from gpaw.setup import types2atomtypes
from gpaw.coulomb import CoulombNEW as Coulomb
from gpaw.mpi import world, rank, serial_comm
from gpaw import GPAW


def get_bf_centers(atoms, basis=None):
    calc = atoms.calc
    if calc is None or isinstance(calc, SinglePointCalculator):
        symbols = atoms.get_chemical_symbols()
        basis_a = types2atomtypes(symbols, basis, 'dzp')
        nao_a = [Basis(symbol, type).nao
                 for symbol, type in zip(symbols, basis_a)]
    else:
        if not calc.initialized:
            calc.initialize(atoms)
        nao_a = [calc.wfs.setups[a].nao for a in range(len(atoms))]
    pos_ic = []
    for pos, nao in zip(atoms.get_positions(), nao_a):
        pos_ic.extend(pos[None].repeat(nao, 0))
    return np.array(pos_ic)


def get_bfi(calc, a_list):
    """basis function indices from a list of atom indices.
       a_list: atom indices
       Use: get_bfi(calc, [0, 4]) gives the functions indices
       corresponding to atom 0 and 4"""
    bfs_list = []
    for a in a_list:
        M = calc.wfs.basis_functions.M_a[a]
        bfs_list += range(M, M + calc.wfs.setups[a].nao)
    return bfs_list


def get_bfi2(symbols, basis, a_list):
    """Same as get_bfi, but does not require an LCAO calc"""
    basis = types2atomtypes(symbols, basis, default='dzp')
    bfs_list = []
    i = 0
    for a, symbol in enumerate(symbols):
        nao = Basis(symbol, basis[a]).nao
        if a in a_list:
            bfs_list += range(i, i + nao)
        i += nao
    return bfs_list


def get_mulliken(calc, a_list):
    """Mulliken charges from a list of atom indices (a_list)."""
    Q_a = {}
    for a in a_list:
        Q_a[a] = 0.0
    for kpt in calc.wfs.kpt_u:
        S_MM = calc.wfs.S_qMM[kpt.q]
        nao = S_MM.shape[0]
        rho_MM = np.empty((nao, nao), calc.wfs.dtype)
        calc.wfs.calculate_density_matrix(kpt.f_n, kpt.C_nM, rho_MM)
        Q_M = np.dot(rho_MM, S_MM).diagonal()
        for a in a_list:
            M1 = calc.wfs.basis_functions.M_a[a]
            M2 = M1 + calc.wfs.setups[a].nao
            Q_a[a] += np.sum(Q_M[M1:M2])
    return Q_a


def get_realspace_hs(h_skmm, s_kmm, bzk_kc, weight_k,
                     R_c=(0, 0, 0), direction='x',
                     symmetry={'enabled': False}):

    from gpaw.symmetry import Symmetry
    from ase.dft.kpoints import get_monkhorst_pack_size_and_offset, \
        monkhorst_pack

    if symmetry['point_group']:
        raise NotImplementedError('Point group symmetry not implemented')

    nspins, nk, nbf = h_skmm.shape[:3]
    dir = 'xyz'.index(direction)
    transverse_dirs = np.delete([0, 1, 2], [dir])
    dtype = float
    if len(bzk_kc) > 1 or np.any(bzk_kc[0] != [0, 0, 0]):
        dtype = complex

    kpts_grid, kpts_shift = get_monkhorst_pack_size_and_offset(bzk_kc)

    # kpts in the transport direction
    nkpts_p = kpts_grid[dir]
    bzk_p_kc = monkhorst_pack((nkpts_p, 1, 1))[:, 0] + kpts_shift[dir]
    weight_p_k = 1. / nkpts_p

    # kpts in the transverse directions
    offset = np.zeros((3,))
    offset[:len(transverse_dirs)] = kpts_shift[transverse_dirs]
    bzk_t_kc = monkhorst_pack(
        tuple(kpts_grid[transverse_dirs]) + (1, )) + offset
    if 'time_reversal' not in symmetry:
        symmetry['time_reversal'] = True
    if symmetry['time_reversal']:
        # XXX a somewhat ugly hack:
        # By default GPAW reduces inversion sym in the z direction. The steps
        # below assure reduction in the transverse dirs.
        # For now this part seems to do the job, but it may be written
        # in a smarter way in the future.
        symmetry = Symmetry([1], np.eye(3))
        symmetry.prune_symmetries_atoms(np.zeros((1, 3)))
        ibzk_kc, ibzweight_k = symmetry.reduce(bzk_kc)[:2]
        ibzk_t_kc, weights_t_k = symmetry.reduce(bzk_t_kc)[:2]
        ibzk_t_kc = ibzk_t_kc[:, :2]
        nkpts_t = len(ibzk_t_kc)
    else:
        ibzk_kc = bzk_kc.copy()
        ibzk_t_kc = bzk_t_kc
        ibzk_t_kc = ibzk_t_kc[:, :2]
        nkpts_t = len(bzk_t_kc)
        weights_t_k = [1. / nkpts_t for k in range(nkpts_t)]

    h_skii = np.zeros((nspins, nkpts_t, nbf, nbf), dtype)
    if s_kmm is not None:
        s_kii = np.zeros((nkpts_t, nbf, nbf), dtype)

    tol = 7
    for j, k_t in enumerate(ibzk_t_kc):
        for k_p in bzk_p_kc:
            k = np.zeros((3,))
            k[dir] = k_p
            k[transverse_dirs] = k_t
            kpoint_list = [list(np.round(k_kc, tol)) for k_kc in ibzk_kc]
            if list(np.round(k, tol)) not in kpoint_list:
                k = -k  # inversion
                index = kpoint_list.index(list(np.round(k, tol)))
                h = h_skmm[:, index].conjugate()
                if s_kmm is not None:
                    s = s_kmm[index].conjugate()
                k = -k
            else:  # kpoint in the ibz
                index = kpoint_list.index(list(np.round(k, tol)))
                h = h_skmm[:, index]
                if s_kmm is not None:
                    s = s_kmm[index]

            c_k = np.exp(2.j * np.pi * np.dot(k, R_c)) * weight_p_k
            h_skii[:, j] += c_k * h
            if s_kmm is not None:
                s_kii[j] += c_k * s

    if s_kmm is None:
        return ibzk_t_kc, weights_t_k, h_skii
    else:
        return ibzk_t_kc, weights_t_k, h_skii, s_kii


def remove_pbc(atoms, h, s=None, d=0, centers_ic=None, cutoff=None):
    if h.ndim > 2:
        raise KeyError('You have to run remove_pbc for each '
                       'spin/kpoint separately.')
    L = atoms.cell[d, d]
    nao = len(h)
    dtype = h.dtype
    if centers_ic is None:
        centers_ic = get_bf_centers(atoms)  # requires an attached LCAO calc
    ni = len(centers_ic)
    if nao != ni:
        assert nao == 2 * ni
        centers_ic = np.vstack((centers_ic, centers_ic))
        centers_ic[ni:, d] += L
        if cutoff is None:
            cutoff = L - 1e-3
    elif cutoff is None:
        cutoff = 0.5 * L - 1e-3
    pos_i = centers_ic[:, d]
    for i in range(nao):
        dpos_i = abs(pos_i - pos_i[i])
        mask_i = (dpos_i < cutoff).astype(dtype)
        h[i, :] *= mask_i
        h[:, i] *= mask_i
        if s is not None:
            s[i, :] *= mask_i
            s[:, i] *= mask_i


def dump_hamiltonian(filename, atoms, direction=None, Ef=None):
    h_skmm, s_kmm = get_lcao_hamiltonian(atoms.calc)
    if direction is not None:
        d = 'xyz'.index(direction)
        for s in range(atoms.calc.nspins):
            for k in range(atoms.calc.nkpts):
                if s == 0:
                    remove_pbc(atoms, h_skmm[s, k], s_kmm[k], d)
                else:
                    remove_pbc(atoms, h_skmm[s, k], None, d)

    if atoms.calc.master:
        with open(filename, 'wb') as fd:
            pickle.dump((h_skmm, s_kmm), fd, 2)
            atoms_data = {'cell': atoms.cell, 'positions': atoms.positions,
                          'numbers': atoms.numbers, 'pbc': atoms.pbc}

            pickle.dump(atoms_data, fd, 2)
            calc_data = {'weight_k': atoms.calc.weight_k,
                         'ibzk_kc': atoms.calc.ibzk_kc}

            pickle.dump(calc_data, fd, 2)

    world.barrier()


def dump_hamiltonian_parallel(filename, atoms, direction=None, Ef=None):
    raise DeprecationWarning(
        'Please use dump_hamiltonian_and_overlap() instead.')


def dump_hamiltonian_and_overlap(filename, atoms, direction=None):
    """
        Dump the lcao representation of H and S to file(s) beginning
        with filename. If direction is x, y or z, the periodic boundary
        conditions will be removed in the specified direction.

        Note:
        H and S are parallized over spin and k-points and
        is for now dumped into a number of pickle files. This
        may be changed into a dump to a single file in the future.

    """
    if direction is not None:
        d = 'xyz'.index(direction)

    calc = atoms.calc
    wfs = calc.wfs
    nao = wfs.setups.nao
    nq = len(wfs.kpt_u) // wfs.nspins
    H_qMM = np.empty((wfs.nspins, nq, nao, nao), wfs.dtype)
    calc_data = {'k_q': {},
                 'skpt_qc': np.empty((nq, 3)),
                 'weight_q': np.empty(nq)}

    S_qMM = wfs.S_qMM

    for kpt in wfs.kpt_u:
        calc_data['skpt_qc'][kpt.q] = calc.wfs.kd.ibzk_kc[kpt.k]
        calc_data['weight_q'][kpt.q] = calc.wfs.kd.weight_k[kpt.k]
        calc_data['k_q'][kpt.q] = kpt.k
        H_MM = wfs.eigensolver.calculate_hamiltonian_matrix(calc.hamiltonian,
                                                            wfs,
                                                            kpt)

        H_qMM[kpt.s, kpt.q] = H_MM

        tri2full(H_qMM[kpt.s, kpt.q])
        if kpt.s == 0:
            tri2full(S_qMM[kpt.q])
            if direction is not None:
                remove_pbc(atoms, H_qMM[kpt.s, kpt.q], S_qMM[kpt.q], d)
        else:
            if direction is not None:
                remove_pbc(atoms, H_qMM[kpt.s, kpt.q], None, d)

    if wfs.gd.comm.rank == 0:
        with open(filename + '%i.pckl' % wfs.kd.comm.rank, 'wb') as fd:
            H_qMM *= Ha
            pickle.dump((H_qMM, S_qMM), fd, 2)
            pickle.dump(calc_data, fd, 2)


def get_lcao_hamiltonian(calc):
    """Return H_skMM, S_kMM on master, (None, None) on slaves. H is in eV."""
    if calc.wfs.S_qMM is None:
        calc.wfs.set_positions(calc.spos_ac)
    dtype = calc.wfs.dtype
    NM = calc.wfs.eigensolver.nao
    Nk = calc.wfs.kd.nibzkpts
    Ns = calc.wfs.nspins

    S_kMM = np.zeros((Nk, NM, NM), dtype)
    H_skMM = np.zeros((Ns, Nk, NM, NM), dtype)
    for kpt in calc.wfs.kpt_u:
        H_MM = calc.wfs.eigensolver.calculate_hamiltonian_matrix(
            calc.hamiltonian, calc.wfs, kpt)
        if kpt.s == 0:
            S_kMM[kpt.k] = calc.wfs.S_qMM[kpt.q]
            tri2full(S_kMM[kpt.k])
        H_skMM[kpt.s, kpt.k] = H_MM * Ha
        tri2full(H_skMM[kpt.s, kpt.k])
    calc.wfs.kd.comm.sum(S_kMM, 0)
    calc.wfs.kd.comm.sum(H_skMM, 0)
    if rank == 0:
        return H_skMM, S_kMM
    else:
        return None, None


def get_lead_lcao_hamiltonian(calc, direction='x'):
    H_skMM, S_kMM = get_lcao_hamiltonian(calc)
    if rank == 0:
        return lead_kspace2realspace(H_skMM, S_kMM,
                                     bzk_kc=calc.wfs.kd.bzk_kc,
                                     weight_k=calc.wfs.kd.weight_k,
                                     direction=direction,
                                     symmetry=calc.parameters.symmetry)
    else:
        return None, None, None, None


def lead_kspace2realspace(h_skmm, s_kmm, bzk_kc, weight_k, direction='x',
                          symmetry={'point_group': False}):
    """Convert a k-dependent Hamiltonian to tight-binding onsite and coupling.

    For each transverse k-point:
    Convert k-dependent (in transport direction) Hamiltonian
    representing a lead to a real space tight-binding Hamiltonian
    of double size representing two principal layers and the coupling between.
    """

    dir = 'xyz'.index(direction)
    if symmetry['point_group']:
        raise NotImplementedError

    R_c = [0, 0, 0]
    ibz_t_kc, weight_t_k, h_skii, s_kii = get_realspace_hs(
        h_skmm, s_kmm, bzk_kc, weight_k, R_c, direction, symmetry)

    R_c[dir] = 1
    h_skij, s_kij = get_realspace_hs(
        h_skmm, s_kmm, bzk_kc, weight_k, R_c, direction, symmetry)[-2:]

    nspins, nk, nbf = h_skii.shape[:-1]

    h_skmm = np.zeros((nspins, nk, 2 * nbf, 2 * nbf), h_skii.dtype)
    s_kmm = np.zeros((nk, 2 * nbf, 2 * nbf), h_skii.dtype)
    h_skmm[:, :, :nbf, :nbf] = h_skmm[:, :, nbf:, nbf:] = h_skii
    h_skmm[:, :, :nbf, nbf:] = h_skij
    h_skmm[:, :, nbf:, :nbf] = h_skij.swapaxes(2, 3).conj()

    s_kmm[:, :nbf, :nbf] = s_kmm[:, nbf:, nbf:] = s_kii
    s_kmm[:, :nbf, nbf:] = s_kij
    s_kmm[:, nbf:, :nbf] = s_kij.swapaxes(1, 2).conj()

    return ibz_t_kc, weight_t_k, h_skmm, s_kmm


def zeta_pol(basis):
    """Get number of zeta func. and polarization func. indices in Basis."""
    zeta = []
    pol = []
    for bf in basis.bf_j:
        if 'polarization' in bf.type:
            pol.append(2 * bf.l + 1)
        else:
            zeta.append(2 * bf.l + 1)
    zeta = sum(zeta)
    pol = sum(pol)
    assert zeta + pol == basis.nao
    return zeta, pol


def basis_subset(symbol, largebasis, smallbasis):
    """Title.

    Determine which basis function indices from ``largebasis`` are also
    present in smallbasis.
    """
    blarge = Basis(symbol, largebasis)
    zeta_large, pol_large = zeta_pol(blarge)

    bsmall = Basis(symbol, smallbasis)
    zeta_small, pol_small = zeta_pol(bsmall)

    assert zeta_small <= zeta_large
    assert pol_small <= pol_large

    insmall = np.zeros(blarge.nao, bool)
    insmall[:zeta_small] = True
    insmall[zeta_large:zeta_large + pol_small] = True
    return insmall


def basis_subset2(symbols, largebasis='dzp', smallbasis='sz'):
    """Same as basis_subset, but for an entire list of atoms."""
    largebasis = types2atomtypes(symbols, largebasis, default='dzp')
    smallbasis = types2atomtypes(symbols, smallbasis, default='sz')
    mask = []
    for symbol, large, small in zip(symbols, largebasis, smallbasis):
        mask.extend(basis_subset(symbol, large, small))
    return np.asarray(mask, bool)


def collect_orbitals(a_xo, coords, root=0):
    """Collect array distributed over orbitals to root-CPU.

    Input matrix has last axis distributed amongst CPUs,
    return is None on slaves, and the collected array on root.

    The distribution can be uneven amongst CPUs. The list coords gives the
    number of values for each CPU.
    """
    a_xo = np.ascontiguousarray(a_xo)
    if world.size == 1:
        return a_xo

    # All slaves send their piece to ``root``:
    # There can be several sends before the corresponding receives
    # are posted, so use syncronous send here
    if world.rank != root:
        world.ssend(a_xo, root, 112)
        return None

    # On root, put the subdomains from the slaves into the big array
    # for the whole domain on root:
    xshape = a_xo.shape[:-1]
    Norb2 = sum(coords)  # total number of orbital indices
    a_xO = np.empty(xshape + (Norb2,), a_xo.dtype)
    o = 0
    for rankx, norb in enumerate(coords):
        if rankx != root:
            tmp_xo = np.empty(xshape + (norb,), a_xo.dtype)
            world.receive(tmp_xo, rankx, 112)
            a_xO[..., o:o + norb] = tmp_xo
        else:
            a_xO[..., o:o + norb] = a_xo
        o += norb
    return a_xO


def makeU(gpwfile='grid.gpw', orbitalfile='w_wG__P_awi.pckl',
          rotationfile='eps_q__U_pq.pckl', tolerance=1e-5,
          writeoptimizedpairs=False, dppname='D_pp.pckl', S_w=None):

    # S_w: None or diagonal of overlap matrix. In the latter case
    # the optimized and truncated pair orbitals are obtained from
    # normalized (to 1) orbitals.
    #
    # Tolerance is used for truncation of optimized pairorbitals
    # calc = GPAW(gpwfile, txt=None)
    from gpaw import GPAW
    from gpaw.mpi import world
    calc = GPAW(gpwfile, txt='pairorb.txt')  # XXX
    gd = calc.wfs.gd
    setups = calc.wfs.setups
    myatoms = calc.density.D_asp.keys()
    del calc

    # Load orbitals on master and distribute to slaves
    if world.rank == 0:
        with open(orbitalfile, 'rb') as fd:
            wglobal_wG, P_awi = pickle.load(fd)
        Nw = len(wglobal_wG)
        print('Estimated total (serial) mem usage: %0.3f GB' % (
            np.prod(gd.N_c) * Nw**2 * 8 / 1024.**3))
    else:
        wglobal_wG = None
        Nw = 0
    Nw = gd.comm.sum(Nw)  # distribute Nw to all nodes
    w_wG = gd.empty(n=Nw)
    gd.distribute(wglobal_wG, w_wG)
    del wglobal_wG

    # Make pairorbitals
    f_pG = gd.zeros(n=Nw**2)
    for p, (w1, w2) in enumerate(np.ndindex(Nw, Nw)):
        np.multiply(w_wG[w1], w_wG[w2], f_pG[p])
    del w_wG
    assert f_pG.flags.contiguous
    # Make pairorbital overlap (lower triangle only)
    D_pp = np.zeros((Nw**2, Nw**2))
    rk(gd.dv, f_pG, 0., D_pp)

    # Add atomic corrections to pairorbital overlap
    for a in myatoms:
        if setups[a].type != 'ghost':
            P_pp = np.array([pack(np.outer(P_awi[a][w1], P_awi[a][w2]))
                             for w1, w2 in np.ndindex(Nw, Nw)])
            I4_pp = setups[a].four_phi_integrals()
            A = np.zeros((len(I4_pp), len(P_pp)))
            mmm(1.0, I4_pp, 'N', P_pp, 'T', 0.0, A)
            mmm(1.0, P_pp, 'N', A, 'N', 1.0, D_pp)
            # D_pp += np.dot(P_pp, np.dot(I4_pp, P_pp.T))

    # Summ all contributions to master
    gd.comm.sum(D_pp, 0)

    if world.rank == 0:
        if S_w is not None:
            print('renormalizing pairorb overlap matrix (D_pp)')
            S2 = np.sqrt(S_w)
            for pa, (wa1, wa2) in enumerate(np.ndindex(Nw, Nw)):
                for pb, (wb1, wb2) in enumerate(np.ndindex(Nw, Nw)):
                    D_pp[pa, pb] /= S2[wa1] * S2[wa2] * S2[wb1] * S2[wb2]

        D_pp.dump(dppname)
        # XXX if the diagonalization below (on MASTER only)
        # fails, then one can always restart the stuff
        # below using only the stored D_pp matrix

        # Determine eigenvalues and vectors on master only
        eps_q, U_pq = np.linalg.eigh(D_pp, UPLO='L')
        del D_pp
        indices = np.argsort(-eps_q.real)
        eps_q = np.ascontiguousarray(eps_q.real[indices])
        U_pq = np.ascontiguousarray(U_pq[:, indices])

        # Truncate
        indices = eps_q > tolerance
        U_pq = np.ascontiguousarray(U_pq[:, indices])
        eps_q = np.ascontiguousarray(eps_q[indices])

        # Dump to file
        with open(rotationfile, 'wb') as fd:
            pickle.dump((eps_q, U_pq), fd, 2)

    if writeoptimizedpairs is not False:
        assert world.size == 1  # works in parallel if U and eps are broadcast
        Uisq_qp = (U_pq / np.sqrt(eps_q)).T.copy()
        g_qG = gd.zeros(n=len(eps_q))
        mmmx(1.0, Uisq_qp, 'N', f_pG, 'N', 0.0, g_qG)
        g_qG = gd.collect(g_qG)
        if world.rank == 0:
            P_app = dict([(a, np.array([pack(np.outer(P_wi[w1], P_wi[w2]),
                                             tolerance=1e3)
                                        for w1, w2 in np.ndindex(Nw, Nw)]))
                          for a, P_wi in P_awi.items()])
            P_aqp = dict([(a, np.dot(Uisq_qp, Px_pp))
                          for a, Px_pp in P_app.items()])
            with open(writeoptimizedpairs, 'wb') as fd:
                pickle.dump((g_qG, P_aqp), fd, 2)


def makeV(gpwfile='grid.gpw', orbitalfile='w_wG__P_awi.pckl',
          rotationfile='eps_q__U_pq.pckl', coulombfile='V_qq.pckl',
          log='V_qq.log', fft=False):

    with IOContext() as io:
        log = io.openfile(log)
        _makeV(gpwfile=gpwfile, orbitalfile=orbitalfile,
               rotationfile=rotationfile, coulombfile=coulombfile,
               log=log, fft=fft)


def _makeV(gpwfile, orbitalfile, rotationfile, coulombfile, log, fft):
    # Extract data from files
    calc = GPAW(gpwfile, txt=None, communicator=serial_comm)
    coulomb = Coulomb(calc.wfs.gd, calc.wfs.setups, calc.spos_ac, fft)
    with open(orbitalfile, 'rb') as fd:
        w_wG, P_awi = pickle.load(fd)
    with open(rotationfile, 'rb') as fd:
        eps_q, U_pq = pickle.load(fd)
    del calc

    # Make rotation matrix divided by sqrt of norm
    Nq = len(eps_q)
    Ni = len(w_wG)
    Uisq_iqj = (U_pq / np.sqrt(eps_q)).reshape(
        Ni, Ni, Nq).swapaxes(1, 2).copy()
    del eps_q, U_pq

    # Determine number of opt. pairorb on each cpu
    Ncpu = world.size
    nq, R = divmod(Nq, Ncpu)
    nq_r = nq * np.ones(Ncpu, int)
    if R > 0:
        nq_r[-R:] += 1

    # Determine number of opt. pairorb on this cpu
    nq1 = nq_r[world.rank]
    q1end = nq_r[:world.rank + 1].sum()
    q1start = q1end - nq1
    V_qq = np.zeros((Nq, nq1), float)

    def make_optimized(qstart, qend):
        g_qG = np.zeros((qend - qstart,) + w_wG.shape[1:], float)
        P_aqp = {}
        for a, P_wi in P_awi.items():
            ni = P_wi.shape[1]
            nii = ni * (ni + 1) // 2
            P_aqp[a] = np.zeros((qend - qstart, nii), float)
        for w1, w1_G in enumerate(w_wG):
            U = Uisq_iqj[w1, qstart: qend].copy()
            mmmx(1, U, 'N', w1_G * w_wG, 'N', 1, g_qG)
            for a, P_wi in P_awi.items():
                P_wp = np.array([pack(np.outer(P_wi[w1], P_wi[w2]))
                                for w2 in range(Ni)])
                mmm(1., U, 'N', P_wp, 'N', 1.0, P_aqp[a])
        return g_qG, P_aqp

    g1_qG, P1_aqp = make_optimized(q1start, q1end)
    for block, nq2 in enumerate(nq_r):
        if block == world.rank:
            g2_qG, P2_aqp = g1_qG, P1_aqp
            q2start, q2end = q1start, q1end
        else:
            q2end = nq_r[:block + 1].sum()
            q2start = q2end - nq2
            g2_qG, P2_aqp = make_optimized(q2start, q2end)

        for q1, q2 in np.ndindex(nq1, nq2):
            P1_ap = dict([(a, P_qp[q1]) for a, P_qp in P1_aqp.items()])
            P2_ap = dict([(a, P_qp[q2]) for a, P_qp in P2_aqp.items()])
            V_qq[q2 + q2start, q1] = coulomb.calculate(g1_qG[q1], g2_qG[q2],
                                                       P1_ap, P2_ap)
            if q2 == 0 and world.rank == 0:
                T = localtime()
                log.write(
                    'Block %i/%i is %4.1f percent done at %02i:%02i:%02i\n' %
                    (block + 1, world.size,
                     100.0 * q1 / nq1, T[3], T[4], T[5]))
                log.flush()

    # Collect V_qq array on master node
    if world.rank == 0:
        T = localtime()
        log.write('Starting collect at %02i:%02i:%02i\n' % (
            T[3], T[4], T[5]))
        log.flush()

    V_qq = collect_orbitals(V_qq, coords=nq_r, root=0)
    if world.rank == 0:
        # V can be slightly asymmetric due to numerics
        V_qq = 0.5 * (V_qq + V_qq.T)
        V_qq.dump(coulombfile)

        T = localtime()
        log.write('Finished at %02i:%02i:%02i\n' % (
            T[3], T[4], T[5]))
        log.flush()



from ase import Atom, Atoms
from ase.units import Bohr
from gpaw import GPAW
from gpaw.test import equal


def test_xc_nonselfconsistentLDA(in_tmp_dir):
    a = 7.5 * Bohr
    n = 16
    atoms = Atoms([Atom('He', (0.0, 0.0, 0.0))], cell=(a, a, a), pbc=True)
    calc = GPAW(gpts=(n, n, n), nbands=1, xc='LDA')
    atoms.calc = calc
    e1 = atoms.get_potential_energy()
    e1ref = calc.get_reference_energy()
    de12 = calc.get_xc_difference({'name': 'PBE', 'stencil': 1})
    calc = GPAW(gpts=(n, n, n), nbands=1, xc={'name': 'PBE', 'stencil': 1})
    atoms.calc = calc
    e2 = atoms.get_potential_energy()
    e2ref = calc.get_reference_energy()
    de21 = calc.get_xc_difference('LDA')
    print(e1ref + e1 + de12, e2ref + e2)
    print(e1ref + e1, e2ref + e2 + de21)
    print(de12, de21)
    equal(e1ref + e1 + de12, e2ref + e2, 0.02)
    equal(e1ref + e1, e2ref + e2 + de21, 0.025)

    calc.write('PBE.gpw')

    de21b = GPAW('PBE.gpw').get_xc_difference('LDA')
    print(de21, de21b)
    equal(de21, de21b, 9e-8)

    energy_tolerance = 0.0007
    equal(e1, -0.0961003634812, energy_tolerance)  # svnversion 5252
    equal(e2, -0.0790249564625, energy_tolerance)  # svnversion 5252



import pytest
from ase.build import bulk

from gpaw import GPAW, Mixer
from gpaw.convergence_criteria import Eigenstates


class MyConvergenceCriterion(Eigenstates):
    def __init__(self, tol):
        super().__init__(tol)
        self.history = []

    def get_error(self, context):
        value = super().get_error(context)
        self.history.append(value)
        return value


def run(atoms, method, kwargs):
    conv_tol = 1e-9
    conv = MyConvergenceCriterion(conv_tol)

    kwargs = {
        'convergence': {'custom': [conv]},
        'experimental': {'reuse_wfs_method': method},
        **kwargs}

    calc = GPAW(**kwargs)

    atoms.calc = calc
    E1 = atoms.get_potential_energy()
    assert conv.history[-1] < conv_tol
    niter1 = len(conv.history)
    del conv.history[:]

    atoms.rattle(stdev=0.0001)

    E2 = atoms.get_potential_energy()
    niter2 = len(conv.history)
    reuse_error = conv.history[0]

    # If the energy is exactly or suspiciously close to zero, it's because
    # nothing was done at all (something was cached but shouldn't have been)
    delta_e = abs(E2 - E1)
    assert delta_e > 1e-6, delta_e
    return niter1, niter2, reuse_error


@pytest.mark.later
@pytest.mark.parametrize('mode, reuse_type, max_reuse_error', [
    ('pw', 'paw', 1e-5),
    ('pw', None, 1e-4),
    ('fd', 'paw', 1e-4),
    ('fd', None, 1e-3),
])
def test_reuse_wfs(mode, reuse_type, max_reuse_error):
    """Check that wavefunctions are meaningfully reused.

    For a different modes and parameters, this test asserts that the
    initial wavefunction error in the second scf step is below a
    certain threshold, indicating that we are doing better than if
    we started from scratch."""

    atoms = bulk('Si')
    atoms.rattle(stdev=0.01, seed=17)  # Break symmetry

    kwargs = dict(
        mode=mode,
        kpts=(2, 2, 2),
        xc='PBE',
        mixer=Mixer(0.4, 5, 20.0))

    niter1, niter2, reuse_error = run(
        atoms, reuse_type, kwargs)

    # It should at the very least be faster to do the second step:
    assert niter2 < niter1
    assert reuse_error < max_reuse_error


@pytest.mark.later
def test_reuse_sg15(sg15_hydrogen):
    """Test wfs reuse with sg15.

    As of writing this test, the sg15 pseudopotentials have no pseudo
    partial waves, and therefore, reusing them should have no effect."""
    from ase.build import molecule
    atoms = molecule('H2', vacuum=2.0)

    # If we do not rattle, we will get broken symmetry error:
    atoms.rattle(stdev=.001)

    kwargs = dict(
        mode='pw',
        setups={'H': sg15_hydrogen},
        xc='PBE')

    niter1, niter2, reuse_error = run(atoms, 'paw', kwargs)
    assert niter2 < niter1
    assert reuse_error < 1e-5



"""Different kinds of laser pulses.
"""

import numpy as np

###########################
# Main class
###########################


class LaserField:
    """

    """
    def __init__(self):
        pass

    def strength(self, t):
        return np.array([0.0, 0.0, 0.0])


class CWField(LaserField):
    """
    Continuously oscillating laser field which is switch on linearly.

    Parameters:
      e0   field strength  (in atomic units)
      w    field frequency (in atomic units)
      ts   switch on time  (in atomic units)
    """
    def __init__(self, e0, w, ts):
        # FIXME: use eV, ang and attosec
        #           attosec_to_autime, autime_to_attosec, \
        #           eV_to_aufrequency, aufrequency_to_eV
        self.e0 = e0
        self.w = w
        self.ts = ts

    def strength(self, t):
        if t < self.ts:
            c = self.e0 * (t / self.ts) * np.sin(self.w * t)
        else:
            c = self.e0 * np.sin(self.w * t)
        return np.array([0.0, 0.0, c])



import matplotlib
from matplotlib import pylab
from gpaw.test.big.g2_1.pbe_gpaw_nrel_analyse import tag
from gpaw.test.big.g2_1.pbe_gpaw_nrel_analyse import ref as calc
from gpaw.test.big.g2_1.pbe_nwchem_def2_qzvppd_analyse import ref
from ase.data.g2_1 import data
from ase.build import molecule
from ase.data import atomic_numbers
import heapq
import numpy as np
import os
import warnings
# silence matplotlib.use() warning
warnings.filterwarnings('ignore', r'.*This call to matplotlib\.use.*',)


ann_fontsize = 'small'
label_fontsize = 12


formulas = sorted(ref['ea'].keys())

atoms = []
for f in formulas:
    m = molecule(f, data=data)
    atoms.extend(m.get_chemical_symbols())
atoms = list(set(atoms))
anum = [atomic_numbers[a] for a in atoms]

# sort according to anum
tmpsort = sorted(zip(anum, atoms))
a, atomssorted = zip(*tmpsort)


def containatom(atom, formulas, data):
    molecules = []
    for f in formulas:
        m = molecule(f, data=data)
        if atom in list(set(m.get_chemical_symbols())):
            molecules.append(f)
    return list(set(molecules))


def get_statistics(result, reference):
    skeys = sorted(result.keys())
    res1 = []
    res2 = []
    for k in skeys:
        res1.append(k)
        res2.append(result[k])
    skeys = reference.keys()
    skeys.sort()
    ref1 = []
    ref2 = []
    for k in skeys:
        ref1.append(k)
        ref2.append(reference[k])
    resrm = []
    refrm = []
    # remove result without reference and reference without result
    toremove = list(set(res1).symmetric_difference(set(ref1)))
    for c in toremove:
        if c in res1:
            print(result + ' vs ' + reference + ': ' + result +
                  ' removed ' + c)
            i = res1.index(c)
            res1.pop(i)
            res2.pop(i)
            resrm.append(c)
        if c in ref1:
            print(result + ' vs ' + reference + ': ' + reference +
                  ' removed ' + c)
            i = ref1.index(c)
            ref1.pop(i)
            ref2.pop(i)
            refrm.append(c)
    # calculate result - reference
    diff = np.array(res2) - np.array(ref2)
    all = [(ref1[list(diff).index(e)], e) for e in diff]
    average = np.average(diff)
    absaverage = np.average(abs(diff))
    std = np.std(diff)
    nlargest = heapq.nlargest(4, abs(diff))
    # n largest errors
    largest = []
    for abse in nlargest:
        c = ref1[list(abs(diff)).index(abse)]
        e = diff[ref1.index(c)]
        largest.append((c, e))
    return average, absaverage, std, largest, all, resrm, refrm

# prepare plot


vs = [
    (calc['ea'].copy(), ref['ea'].copy(), 'G2_1')]

for atom in atoms:
    calc_atom = {}
    for a in containatom(atom, formulas, data):
        calc_atom.update({a: calc['ea'][a]})
    ref_atom = {}
    for a in containatom(atom, formulas, data):
        ref_atom.update({a: ref['ea'][a]})
    vs.append((calc_atom, ref_atom, atom))

no = []
average = []
absaverage = []
std = []
largest = []
all = []
resrm = []
refrm = []
labelsbase = []
labels = []
for n, results in enumerate(vs):
    label = results[2]
    labelsbase.append(label)
    labels.append(label)
    a, absa, s, l, al, rs, rf = get_statistics(results[0], results[1])
    no.append(n)
    average.append(a)
    absaverage.append(absa)
    std.append(s)
    largest.append(l)
    resrm.append(rs)
    refrm.append(rf)
    all.append(al)
# print no
# print average
# print absaverage
# print std
# print largest


def plot(xdata, ydata, std,
         title,
         xlabel, ylabel,
         label, color, alpha,
         miny, maxy,
         num=1,
         ):
    import matplotlib
    # matplotlib.use('Agg')
    import pylab
    import matplotlib.font_manager

    # all goes to figure num
    pylab.figure(num=num, figsize=(9.5, 9))
    pylab.gca().set_position([0.10, 0.20, 0.85, 0.60])
    # let the plot have fixed y-axis scale
    # ywindow = maxy - miny
    # pylab.gca().set_ylim(miny, maxy+ywindow/5.0)
    pylab.gca().set_ylim(miny, maxy)
    # pylab.plot(xdata, ydata, 'b.', label=label, color=color)
    # pylab.plot(xdata, ydata, 'b-', label='_nolegend_', color=color)
    pylab.bar(xdata, ydata, 0.9, label=label, color=color, alpha=alpha)
    t = pylab.title(title)
    t.set_y(1.05)
    pylab.xlabel(xlabel)
    pylab.ylabel(ylabel)
    prop = matplotlib.font_manager.FontProperties(size=12)
    leg = pylab.legend(loc='upper right', fancybox=True, prop=prop)
    leg.get_frame().set_alpha(0.5)


def plot_save(directory_name, out_prefix):
    from os.path import exists
    assert exists(directory_name)
    import pylab

    pylab.savefig(directory_name + os.path.sep + out_prefix + '.png',
                  bbox_inches='tight')


matplotlib.use('Agg')
# print scaling results

num = 1
miny = -0.3
maxy = 0.3

plot(
    no, average, std,
    'Atomization energy: E(atoms) - E(molecule)',
    'Subsets contain',
    'Calculated - Reference [eV]',
    'Average (avg)',
    'blue',
    0.5,
    miny, maxy,
    num=1)
plot(
    no, absaverage, std,
    'Atomization energy: E(atoms) - E(molecule)',
    'Subsets contain',
    'Calculated - Reference [eV]',
    'Average absolute (abs)',
    'green',
    0.2,
    miny, maxy,
    num=1)
zero = [0.0 for i in range(len(no))]
pylab.plot(no, zero, 'k-', label='_nolegend_')
ay1 = pylab.gca()
ay1.xaxis.set_ticks([n + 0.5 for n in no])
ay1.xaxis.set_ticklabels(labels)
ay1.yaxis.set_minor_locator(matplotlib.ticker.AutoMinorLocator())

for label in ay1.get_xticklabels() + ay1.get_yticklabels():
    label.set_fontsize(label_fontsize)


def errorslocation(n, n1):
    return (n + 0.1, 0.25 - 0.05 * n1)


def formulaslocation(n, n1):
    return (n + 0.3, - 0.10 - 0.05 * n1)


for n in range(len(no)):
    label = ''
    for n1, v in enumerate(['average', 'absaverage', 'std']):
        l = {
            'average': '\navg\n',
            'absaverage': '\nabs\n',
            'std': '\nstd\n'}[v]
        value = {
            'average': average,
            'absaverage': absaverage,
            'std': std}[v]
        label += l + ' ' + str(round(value[n], 3)) + '\n'
    pylab.annotate(label,
                   xy=(n + 0.0, 0.0),
                   xytext=errorslocation(n, n1),
                   arrowprops=None,
                   horizontalalignment='left', verticalalignment='center',
                   fontsize=ann_fontsize)

# plot compounds with largest errors
for n, l in enumerate(largest):
    for n1, (c, e) in enumerate(l):
        name = c + '\n'
        # matplotlib.pyparsing.ParseFatalException: Expected end of math '$'
        # $\rm{SiH}2_\rm{s}3\rm{B}1\rm{d}$ (at char 0), (line:1, col:1)
        name = name.replace('\\rm', '')
        label = name + ' ' + str(round(e, 2))
        pylab.annotate(label,
                       xy=(n + 0.05, e),
                       xytext=formulaslocation(n, n1),
                       arrowprops=dict(width=0.05, headwidth=5.0,
                                       facecolor='black', shrink=1.00),
                       horizontalalignment='left', verticalalignment='center',
                       fontsize=ann_fontsize)
# pylab.show()
plot_save(".", tag + '_ea_vs')
# pylab.close(1)



from math import nan
from gpaw.yml import comment


class PoissonSolver:
    def solve(self,
              vHt,
              rhot) -> float:
        raise NotImplementedError


class PoissonSolverWrapper(PoissonSolver):
    def __init__(self, solver):
        self.description = solver.get_description()
        self.solver = solver

    def __str__(self):
        return comment(self.description)

    def solve(self,
              vHt,
              rhot) -> float:
        self.solver.solve(vHt.data, rhot.data)
        return nan



import os
import fnmatch
from io import BytesIO
import tarfile
import re
from urllib.request import urlopen
import ssl


sources = [('gpaw', 'official GPAW setups releases [default]'),
           ('sg15', 'SG15 pseudopotentials'),
           ('basis', 'basis sets for LCAO mode'),
           ('test', 'small file for testing this script')]

names = [r for r, d in sources]

baseurls = {
    'gpaw': 'https://wiki.fysik.dtu.dk/gpaw/_sources/setups/setups.rst.txt',
    'sg15': 'http://www.quantum-simulation.org/potentials/sg15_oncv/',
    'basis': 'https://wiki.fysik.dtu.dk/gpaw-files/',
    'test': 'https://wiki.fysik.dtu.dk/gpaw-files/'}


notfound_msg = """\
For some reason the files were not found.

Perhaps this script is out of date, and the data is no longer
available at the expected URL:

  {url}

Or maybe there it is just a temporary problem or timeout.  Please try
again, or rummage around the GPAW web page until a solution is found.
Writing e-mails to gpaw-users@listserv.fysik.dtu.dk or reporting
an issue on https://gitlab.com/gpaw/gpaw/issues is also
likely to help."""


def urlopen_nocertcheck(src):
    """Open a URL on a server without checking the certificate.

    Some data is read from a DTU server with a self-signed
    certificate.  That causes trouble on some machines.
    """

    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    return urlopen(src, context=ctx)


class CLICommand:
    """Install PAW datasets, pseudopotential or basis sets.

    Without a directory, show available setups and GPAW
    setup paths.

    With a directory, download and install gpaw-setups into
    INSTALLDIR/[setups-package-name-and-version].
    """

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument
        add('directory', nargs='?')
        add('--version',
            help='download VERSION of package.  '
            'Run without arguments to display a list of versions.  '
            'VERSION can be the full URL or a part such as  '
            '\'0.8\' or \'0.6.6300\'')
        add('--tarball', metavar='FILE',
            help='unpack and install from local tarball FILE '
            'instead of downloading')
        add('--list-all', action='store_true',
            help='list packages from all sources')
        for name, help in sources:
            add('--' + name, action='store_const',
                const=name, dest='source',
                help=help)
        add('--register', action='store_true',
            help='run non-interactively and register install path in '
            'GPAW setup search paths.  This is done by adding lines to '
            '~/.gpaw/rc.py')
        add('--no-register', action='store_true',
            help='run non-interactively and do not register install path in '
            'GPAW setup search paths')

    @staticmethod
    def run(args, parser):
        main(args, parser)


def main(args, parser):
    if args.source is None:
        args.source = sources[0][0]

    if args.register and args.no_register:
        parser.error('Conflicting options specified on whether to register '
                     'setup install paths in configuration file.  Try not '
                     'specifying some options.')

    # The sg15 file is a tarbomb.  We will later defuse it by untarring
    # into a subdirectory, so we don't leave a ghastly mess on the
    # unsuspecting user's system.

    if not args.tarball:
        if args.list_all:
            urls = []
            for source in names:
                urls1 = get_urls(source)
                urls.extend(urls1)
        else:
            urls = get_urls(args.source)

        def print_urls(urls, marked=None):
            for url in urls:
                pageurl, fname = url.rsplit('/', 1)
                if url == marked:
                    marking = ' [*]'
                else:
                    marking = '    '
                print(' %s %s' % (marking, url))

        if len(urls) == 0:
            url = baseurls[args.source]
            parser.error(notfound_msg.format(url=url))

        if args.version:
            matching_urls = [url for url in urls if args.version in url]
            if len(matching_urls) > 1:
                parser.error('More than one setup file matches version "%s":\n'
                             '%s' % (args.version, '\n'.join(matching_urls)))
            elif len(matching_urls) == 0:
                parser.error('\nNo setup matched the specified version "%s".\n'
                             'Available setups are:\n'
                             '%s' % (args.version, '\n'.join(urls)))
            assert len(matching_urls) == 1
            url = matching_urls[0]
        else:
            url = urls[0]

        print('Available setups and pseudopotentials')
        print_urls(urls, url)
        print()

    if not args.directory:
        print_setups_info(parser)
        print()
        print('Run gpaw install-data DIR to install newest setups into DIR.')
        print('Run gpaw install-data DIR --version=VERSION to install VERSION '
              '(from above).')
        print('See gpaw install-data --help for more info.')
        raise SystemExit

    targetpath = args.directory

    if args.tarball:
        print('Reading local tarball %s' % args.tarball)
        targzfile = tarfile.open(args.tarball)
        tarfname = args.tarball
    else:
        tarfname = url.rsplit('/', 1)[1]
        print('Selected %s.  Downloading...' % tarfname)
        response = urlopen_nocertcheck(url)
        targzfile = tarfile.open(fileobj=BytesIO(response.read()))

    if not os.path.exists(targetpath):
        os.makedirs(targetpath)

    assert tarfname.endswith('.tar.gz')
    setup_dirname = tarfname.rsplit('.', 2)[0]  # remove .tar.gz ending
    setup_path = os.path.abspath(os.path.join(targetpath, setup_dirname))
    if tarfname.startswith('sg15'):
        # Defuse tarbomb
        if not os.path.isdir(setup_path):
            os.mkdir(setup_path)
        targetpath = os.path.join(targetpath, setup_dirname)

    print('Extracting tarball into %s' % targetpath)
    targzfile.extractall(targetpath)
    assert os.path.isdir(setup_path)
    print('Setups installed into %s.' % setup_path)

    # Okay, now we have to maybe edit people's rc files.
    rcfiledir = os.path.join(os.environ['HOME'], '.gpaw')
    rcfilepath = os.path.join(rcfiledir, 'rc.py')

    # We could do all this by importing the rcfile as well and checking
    # whether things are okay or not.
    rcline = "setup_paths.insert(0, '%s')" % setup_path

    # Run interactive mode unless someone specified a flag requiring otherwise
    interactive_mode = not (args.register or args.no_register)

    register_path = False

    if interactive_mode:
        answer = input('Register this setup path in %s? [y/n] ' % rcfilepath)
        if answer.lower() in ['y', 'yes']:
            register_path = True
        elif answer.lower() in ['n', 'no']:
            print('As you wish.')
        else:
            print('What do you mean by "%s"?  Assuming "n".' % answer)
    else:
        if args.register:
            assert not args.no_register
            register_path = True
        else:
            assert args.no_register

    if register_path:
        # First we create the file
        if not os.path.exists(rcfiledir):
            os.makedirs(rcfiledir)
        if not os.path.exists(rcfilepath):
            tmpfd = open(rcfilepath, 'w')  # Just create empty file
            tmpfd.close()

        for line in open(rcfilepath):
            if line.startswith(rcline):
                print('It looks like the path is already registered in %s.'
                      % rcfilepath)
                print('File will not be modified at this time.')
                break
        else:
            rcfd = open(rcfilepath, 'a')
            print(rcline, file=rcfd)
            print('Setup path registered in %s.' % rcfilepath)
            # Need to explicitly flush/close the file so print_setups_info
            # sees the change in rc.py
            rcfd.close()

            print_setups_info(parser)
    else:
        print('You can manually register the setups by adding the')
        print('following line to %s:' % rcfilepath)
        print()
        print(rcline)
        print()
    print('Installation complete.')


def get_urls(source):
    page = baseurls[source]
    if source == 'gpaw':
        response = urlopen_nocertcheck(page)
        pattern = 'https://wiki.fysik.dtu.dk/gpaw-files/gpaw-setups-*.tar.gz'
        lines = (line.strip().decode() for line in response)
        urls = [line for line in lines if fnmatch.fnmatch(line, pattern)]

    elif source == 'sg15':
        response = urlopen_nocertcheck(page)

        # We want sg15_oncv_2015-10-07.tar.gz, but they may upload
        # newer files, too.
        pattern = (r'<a\s*href=[^>]+>\s*'
                   r'(sg15_oncv_upf_\d\d\d\d-\d\d-\d\d.tar.gz)'
                   r'\s*</a>')

        txt = response.read().decode('ascii', errors='replace')
        files = re.compile(pattern).findall(txt)
        files.sort(reverse=True)
        urls = [page + fname for fname in files]

    elif source == 'basis':
        files = ['gpaw-basis-NAO-sz+coopt-NGTO-0.9.11271.tar.gz',
                 'gpaw-basis-pvalence-0.9.11271.tar.gz',
                 'gpaw-basis-pvalence-0.9.20000.tar.gz']
        urls = [page + fname for fname in files]

    elif source == 'test':
        urls = [page + 'gpaw-dist-test-source.tar.gz']

    else:
        raise ValueError('Unknown source: %s' % source)

    return urls


def print_setups_info(parser):
    try:
        import gpaw
    except ImportError as e:
        parser.error('Cannot import \'gpaw\'.  GPAW does not appear to be '
                     'installed. %s' % e)

    # The contents of the rc file may have changed.  Thus, we initialize
    # setup_paths again to be sure that everything is as it should be.
    gpaw.setup_paths[:] = []
    gpaw.read_rc_file()
    gpaw.initialize_data_paths()

    npaths = len(gpaw.setup_paths)
    if npaths == 0:
        print('GPAW currently has no setup search paths')
    else:
        print('Current GPAW setup paths in order of search priority:')
        for i, path in enumerate(gpaw.setup_paths):
            print('%4d. %s' % (i + 1, path))



import numpy as np
from abc import ABC, abstractmethod

from ase.units import Hartree

from gpaw.utilities.progressbar import ProgressBar

from gpaw.response import timer
from gpaw.response.kspair import (KohnShamKPointPair,
                                  KohnShamKPointPairExtractor)
from gpaw.response.pw_parallelization import block_partition
from gpaw.response.pair_functions import SingleQPWDescriptor, PairFunction
from gpaw.response.pair_transitions import PairTransitions


class PairFunctionIntegrator(ABC):
    r"""Baseclass for computing pair functions in the Kohn-Sham system.

    The implementation is currently restricted collinear periodic crystals in
    absence of spin-orbit coupling.

    In the Kohn-Sham system, pair functions (see the PairFunciton class for
    further descriptions) can be constructed straight-forwardly as a sum over
    transitions between Kohn-Sham eigenstates at k and k + q,
                  __  __  __                            __
               1  \   \   \                          1  \
    pf(q,z) =  ‾  /   /   /   pf_nks,n'k+qs'(q,z) =  ‾  /  pf_T(q,z)
               V  ‾‾  ‾‾  ‾‾                         V  ‾‾
                  k  n,n' s,s'                          T

    where V is the crystal volume and T is a composite index encoding all
    relevant transitions:

    T: (n, k, s) -> (n', k + q, s')

    The sum over transitions can be split into two steps: (1) an integral over
    k-points k inside the 1st Brillouin Zone and (2) a sum over band and spin
    transitions t:

    t (composite transition index): (n, s) -> (n', s')
                  __                 __  __                  __
               1  \               1  \   \                1  \
    pf(q,z) =  ‾  /  pf_T(q,z) =  ‾  /   /  pf_kt(q,z) =  ‾  /  (...)_k
               V  ‾‾              V  ‾‾  ‾‾               V  ‾‾
                  T                  k   t                   k

    In the code, the k-point integral is handled by a KPointPairIntegral
    object, while the sum over band and spin transitions t is carried out in
    the self.add_integrand() method, which also defines the specific pair
    function in question.

    KPointPairIntegral:
       __
    1  \
    ‾  /  (...)_k
    V  ‾‾
       k

    self.add_integrand():
                __                __   __
                \                 \    \
    (...)_k  =  /  pf_kt(q,z)  =  /    /   pf_nks,n'k+qs'(q,z)
                ‾‾                ‾‾   ‾‾
                t                 n,n' s,s'

    In practise, the integration is carried out by letting the
    KPointPairIntegral extract individual KohnShamKPointPair objects, which
    contain all relevant information about the Kohn-Sham eigenstates at k and
    k + q for a number of specified spin and band transitions t.
    KPointPairIntegral.weighted_kpoint_pairs() generates these kptpairs along
    with their integral weights such that self._integrate() can construct the
    pair functions in a flexible, yet general manner.
    
    NB: Although it is not a fundamental limitation to pair functions as
    described above, the current implementation is based on a plane-wave
    represenation of spatial coordinates. This means that symmetries are
    analyzed with a plane-wave basis in mind, leaving room for further
    generalization in the future.
    """

    def __init__(self, gs, context, nblocks=1,
                 disable_point_group=False,
                 disable_time_reversal=False,
                 disable_non_symmorphic=True):
        """Construct the PairFunctionIntegrator

        Parameters
        ----------
        gs : ResponseGroundStateAdapter
        context : ResponseContext
        nblocks : int
            Distribute the pair function into nblocks. Useful when the pair
            function itself becomes a large array (read: memory limiting).
        disable_point_group : bool
            Do not use the point group symmetry operators.
        disable_time_reversal : bool
            Do not use time reversal symmetry.
        disable_non_symmorphic : bool
            Do no use non symmorphic symmetry operators.
        """
        self.gs = gs
        self.context = context

        # Communicators for distribution of memory and work
        (self.blockcomm,
         self.intrablockcomm) = self.create_communicators(nblocks)
        self.nblocks = self.blockcomm.size

        # The KohnShamKPointPairExtractor class handles extraction of k-point
        # pairs from the ground state
        self.kptpair_extractor = KohnShamKPointPairExtractor(
            self.gs, self.context,
            # Distribution of work:
            # t-transitions are distributed through blockcomm,
            # k-points through intrablockcomm.
            transitions_blockcomm=self.blockcomm,
            kpts_blockcomm=self.intrablockcomm)

        # Symmetry flags
        self.disable_point_group = disable_point_group
        self.disable_time_reversal = disable_time_reversal
        self.disable_non_symmorphic = disable_non_symmorphic
        if (disable_time_reversal and disable_point_group
            and disable_non_symmorphic):
            self.disable_symmetries = True
        else:
            self.disable_symmetries = False

    @timer('Integrate pair function')
    def _integrate(self, out: PairFunction, transitions: PairTransitions):
        """In-place pair function integration

        Parameters
        ----------
        out : PairFunction
            Output data structure
        transitions : PairTransitions
            Band and spin transitions to integrate.

        Returns
        -------
        analyzer : PWSymmetryAnalyzer
        """
        # Initialize the plane-wave symmetry analyzer
        analyzer = self.get_pw_symmetry_analyzer(out.qpd)
        
        # Perform the actual integral as a point integral over k-point pairs
        integral = KPointPairPointIntegral(self.kptpair_extractor, analyzer)
        weighted_kptpairs = integral.weighted_kpoint_pairs(transitions)
        pb = ProgressBar(self.context.fd)  # pb with a generator is awkward
        for _, _ in pb.enumerate([None] * integral.ni):
            kptpair, weight = next(weighted_kptpairs)
            if weight is not None:
                assert kptpair is not None
                self.add_integrand(kptpair, weight, out)

        # Sum over the k-points, which have been distributed between processes
        with self.context.timer('Sum over distributed k-points'):
            self.intrablockcomm.sum(out.array)

        # Because the symmetry analyzer is used both to generate the k-point
        # integral domain *and* to symmetrize pair functions after the
        # integration, we have to return it. It would be good to split up these
        # two tasks, so that we don't need to pass the analyzer object around
        # in the code like this...
        return analyzer

    @abstractmethod
    def add_integrand(self, kptpair: KohnShamKPointPair, weight,
                      out: PairFunction):
        """Add the relevant integrand of the outer k-point integral to the
        output data structure 'out', weighted by 'weight' and constructed
        from the provided KohnShamKPointPair 'kptpair'.

        This method effectively defines the pair function in question.
        """

    def create_communicators(self, nblocks):
        """Create MPI communicators to distribute the memory needed to store
        large arrays and parallelize calculations when possible.

        Parameters
        ----------
        nblocks : int
            Separate large arrays into n different blocks. Each process
            allocates memory for the large arrays. By allocating only a
            fraction/block of the total arrays, the memory requirements are
            eased.

        Returns
        -------
        blockcomm : gpaw.mpi.Communicator
            Communicate between processes belonging to different memory blocks.
            In every communicator, there is one process for each block of
            memory, so that all blocks are represented.
            If nblocks < comm.size, there will be size // nblocks different
            processes that allocate memory for the same block of the large
            arrays. Thus, there will be also size // nblocks different block
            communicators, grouping the processes into sets that allocate the
            entire arrays between them.
        intrablockcomm : gpaw.mpi.Communicator
            Communicate between processes belonging to the same memory block.
            There will be size // nblocks processes per memory block.
        """
        comm = self.context.comm
        blockcomm, intrablockcomm = block_partition(comm, nblocks)

        return blockcomm, intrablockcomm

    def get_pw_descriptor(self, q_c, ecut=50, gammacentered=False):
        q_c = np.asarray(q_c, dtype=float)
        ecut = None if ecut is None else ecut / Hartree  # eV to Hartree
        gd = self.gs.gd

        qpd = SingleQPWDescriptor.from_q(q_c, ecut, gd,
                                         gammacentered=gammacentered)

        return qpd

    def get_pw_symmetry_analyzer(self, qpd):
        from gpaw.response.symmetry import PWSymmetryAnalyzer

        return PWSymmetryAnalyzer(
            self.gs.kd, qpd, self.context,
            disable_point_group=self.disable_point_group,
            disable_time_reversal=self.disable_time_reversal,
            disable_non_symmorphic=self.disable_non_symmorphic)

    def get_band_and_spin_transitions(self, spincomponent, nbands=None,
                                      bandsummation='pairwise'):
        """Get band and spin transitions (n, s) -> (n', s') to integrate."""
        nspins = self.gs.nspins
        gsnbands, nocc1, nocc2 = self.get_band_information()

        # Defaults to inclusion of all bands in the ground state calculator
        if nbands is None:
            nbands = gsnbands
        assert nbands <= gsnbands

        transitions = PairTransitions.from_transitions_domain_arguments(
            spincomponent, nbands, nocc1, nocc2, nspins, bandsummation)

        return transitions

    def get_band_information(self):
        """Get information about band occupation."""
        nbands = self.gs.bd.nbands
        nocc1 = self.kptpair_extractor.nocc1
        nocc2 = self.kptpair_extractor.nocc2

        return nbands, nocc1, nocc2

    def get_basic_info_string(self):
        """Get basic information about the ground state and parallelization."""
        nspins = self.gs.nspins
        nbands, nocc1, nocc2 = self.get_band_information()
        nk = self.gs.kd.nbzkpts
        nik = self.gs.kd.nibzkpts

        csize = self.context.comm.size
        knsize = self.intrablockcomm.size
        bsize = self.blockcomm.size

        s = ''

        s += 'The pair function integration is based on a ground state with:\n'
        s += '    Number of spins: %d\n' % nspins
        s += '    Number of bands: %d\n' % nbands
        s += '    Number of completely occupied bands: %d\n' % nocc1
        s += '    Number of partially occupied bands: %d\n' % nocc2
        s += '    Number of kpoints: %d\n' % nk
        s += '    Number of irredicible kpoints: %d\n' % nik
        s += '\n'
        s += 'The pair function integration is performed in parallel with:\n'
        s += '    comm.size: %d\n' % csize
        s += '    intrablockcomm.size: %d\n' % knsize
        s += '    blockcomm.size: %d\n' % bsize

        return s


class KPointPairIntegral(ABC):
    r"""Baseclass for reciprocal space integrals of the first Brillouin Zone,
    where the integrand is a sum over transitions between any number of states
    at the wave vectors k and k + q (referred to as a k-point pair).

    Definition (V is the total crystal hypervolume and D is the dimension of
    the crystal):
       __
    1  \                 1    /
    ‾  /  (...)_k  =  ‾‾‾‾‾‾  |dk (...)_k
    V  ‾‾             (2π)^D  /
       k

    NB: In the current implementation, the dimension is fixed to 3. This is
    sensible for pair functions which are functions of position (such as the
    four-component Kohn-Sham susceptibility tensor), and in most circumstances
    a change in dimensionality can be accomplished simply by adding an extra
    prefactor to the integral elsewhere.
    NB: The current implementation is running on backbone functionality to
    analyze symmetries within a plane wave representation of real-space
    coordinates. This could be generalized further in the future. See the
    PWSymmetryAnalyzer in gpaw.response.symmetry.
    """

    def __init__(self, kptpair_extractor, analyzer):
        """Construct a KPointPairIntegral corresponding to a given q-point.

        Parameters
        ----------
        kptpair_extractor : KohnShamKPointPairExtractor
            Object responsible for extracting all relevant information about
            the k-point pairs from the underlying ground state calculation.
        analyzer : PWSymmetryAnalyzer
            Object responsible for analyzing the symmetries of the q-point in
            question, for which the k-point pair integral is constructed.
        """
        self.gs = kptpair_extractor.gs
        self.kptpair_extractor = kptpair_extractor
        self.q_c = analyzer.qpd.q_c

        # Prepare the k-point pair integral
        bzk_kc, weight_k = self.get_kpoint_domain(analyzer)
        bzk_ipc, weight_i = self.slice_kpoint_domain(bzk_kc, weight_k)
        self._domain = (bzk_ipc, weight_i)
        self.ni = len(weight_i)

    def weighted_kpoint_pairs(self, transitions):
        r"""Generate all k-point pairs in the integral along with their
        integral weights.

        The reciprocal space integral is estimated as the sum over a discrete
        k-point domain. The domain will genererally depend on the integration
        method as well as the symmetry of the crystal.

        Definition:
                                        __
           1    /            ~     1    \   (2π)^D
        ‾‾‾‾‾‾  |dk (...)_k  =  ‾‾‾‾‾‾  /   ‾‾‾‾‾‾ w_kr (...)_kr
        (2π)^D  /               (2π)^D  ‾‾  Nk V0
                                        kr
                                __
                             ~  \
                             =  /  iw_kr (...)_kr
                                ‾‾
                                kr

        The sum over kr denotes the reduced k-point domain specified by the
        integration method (a reduced selection of Nkr points from the ground
        state k-point grid of Nk total points in the entire 1BZ). Each point
        is weighted by its k-point volume in the ground state k-point grid

                      (2π)^D
        kpointvol  =  ‾‾‾‾‾‾,
                      Nk V0

        and an additional individual k-point weight w_kr specific to the
        integration method (V0 denotes the cell volume). Together with the
        integral prefactor, these make up the integral weight

                   1
        iw_kr = ‾‾‾‾‾‾ kpointvol w_kr
                (2π)^D

        Parameters
        ----------
        transitions : PairTransitions
            Band and spin transitions to integrate.
        """
        # Calculate prefactors
        outer_prefactor = 1 / (2 * np.pi)**3
        V = self.crystal_volume()  # V = Nk * V0
        kpointvol = (2 * np.pi)**3 / V
        prefactor = outer_prefactor * kpointvol

        # Generate k-point pairs
        for k_pc, weight in zip(*self._domain):
            if weight is None:
                integral_weight = None
            else:
                integral_weight = prefactor * weight
            kptpair = self.kptpair_extractor.get_kpoint_pairs(
                k_pc, k_pc + self.q_c, transitions)
            yield kptpair, integral_weight

    @abstractmethod
    def get_kpoint_domain(self, analyzer):
        """Use the PWSymmetryAnalyzer to define and weight the k-point domain.

        Returns
        -------
        bzk_kc : np.array
            k-points to integrate in relative coordinates.
        weight_k : np.array
            Integral weight of each k-point in the integral.
        """

    def slice_kpoint_domain(self, bzk_kc, weight_k):
        """When integrating over k-points, slice the domain in pieces with one
        k-point per process each.

        Returns
        -------
        bzk_ipc : nd.array
            k-points (relative) coordinates for each process for each iteration
        """
        comm = self.kptpair_extractor.kpts_blockcomm
        rank, size = comm.rank, comm.size

        nk = bzk_kc.shape[0]
        ni = (nk + size - 1) // size
        bzk_ipc = [bzk_kc[i * size:(i + 1) * size] for i in range(ni)]

        # Extract the weight corresponding to the process' own k-point pair
        weight_ip = [weight_k[i * size:(i + 1) * size] for i in range(ni)]
        weight_i = [None] * len(weight_ip)
        for i, w_p in enumerate(weight_ip):
            if rank in range(len(w_p)):
                weight_i[i] = w_p[rank]

        return bzk_ipc, weight_i

    def crystal_volume(self):
        """Calculate the total crystal volume, V = Nk * V0, corresponding to
        the ground state k-point grid."""
        return self.gs.kd.nbzkpts * self.gs.volume


class KPointPairPointIntegral(KPointPairIntegral):
    r"""Reciprocal space integral of k-point pairs in the first Brillouin Zone,
    estimated as a point integral over all k-points of the ground state k-point
    grid.

    Definition:
                                   __
       1    /           ~     1    \   (2π)^D
    ‾‾‾‾‾‾  |dk (...)_k =  ‾‾‾‾‾‾  /   ‾‾‾‾‾‾ (...)_k
    (2π)^D  /              (2π)^D  ‾‾  Nk V0
                                   k

    """

    def get_kpoint_domain(self, analyzer):
        # Generate k-point domain in relative coordinates
        K_gK = analyzer.group_kpoints()  # What is g? XXX
        bzk_kc = np.array([self.gs.kd.bzk_kc[K_K[0]] for
                           K_K in K_gK])  # Why only K=0? XXX

        # Get the k-point weights from the symmetry analyzer
        weight_k = np.array([analyzer.get_kpoint_weight(k_c)
                             for k_c in bzk_kc])

        return bzk_kc, weight_k



import pytest
from ase import Atoms

from gpaw import GPAW, FermiDirac
from gpaw.test import gen
from gpaw.xas import XAS, RecursionMethod


@pytest.mark.later
def test_corehole_si(in_tmp_dir, add_cwd_to_setup_paths):
    # Generate setup for oxygen with half a core-hole:
    gen('Si', name='hch1s', corehole=(1, 0, 0.5), gpernode=30, write_xml=True)
    a = 2.6
    si = Atoms('Si', cell=(a, a, a), pbc=True)

    import numpy as np
    calc = GPAW(nbands=None,
                h=0.25,
                occupations=FermiDirac(width=0.05),
                setups='hch1s',
                convergence={'maximum iterations': 1})
    si.calc = calc
    _ = si.get_potential_energy()
    calc.write('si.gpw')

    # restart from file
    calc = GPAW('si.gpw')

    import gpaw.mpi as mpi
    if mpi.size == 1:
        xas = XAS(calc)
        x, y = xas.get_spectra()
    else:
        x = np.linspace(0, 10, 50)

    k = 2
    calc = calc.new(kpts=(k, k, k))
    calc.initialize(si)
    calc.set_positions(si)
    assert calc.wfs.dtype == complex

    r = RecursionMethod(calc)
    r.run(40)
    if mpi.size == 1:
        z = r.get_spectra(x)

    if 0:
        import pylab as p
        p.plot(x, y[0])
        p.plot(x, sum(y))
        p.plot(x, z[0])
        p.show()

    # 2p corehole
    s = gen('Si', name='hch2p', corehole=(2, 1, 0.5), gpernode=30)
    calc = GPAW(nbands=None,
                h=0.25,
                occupations=FermiDirac(width=0.05),
                setups={0: s})
    si.calc = calc

    def stopcalc():
        calc.scf.converged = True

    calc.attach(stopcalc, 1)
    _ = si.get_potential_energy()



from __future__ import annotations

from types import ModuleType
from typing import TYPE_CHECKING, Generic, TypeVar

import gpaw.fftw as fftw
import numpy as np
from ase.io.ulm import NDArrayReader
from gpaw.core.domain import Domain
from gpaw.core.matrix import Matrix
from gpaw.mpi import MPIComm
from gpaw.typing import Array1D

if TYPE_CHECKING:
    from gpaw.core.uniform_grid import UniformGridFunctions, UniformGrid

from gpaw.new import prod

DomainType = TypeVar('DomainType', bound=Domain)


class DistributedArrays(Generic[DomainType]):
    desc: DomainType

    def __init__(self,
                 dims: int | tuple[int, ...],
                 myshape: tuple[int, ...],
                 comm: MPIComm,
                 domain_comm: MPIComm,
                 data: np.ndarray | None,
                 dv: float,
                 dtype,
                 xp=None):
        self.myshape = myshape
        self.comm = comm
        self.domain_comm = domain_comm
        self.dv = dv

        # convert int to tuple:
        self.dims = dims if isinstance(dims, tuple) else (dims,)

        if self.dims:
            mydims0 = (self.dims[0] + comm.size - 1) // comm.size
            d1 = min(comm.rank * mydims0, self.dims[0])
            d2 = min((comm.rank + 1) * mydims0, self.dims[0])
            mydims0 = d2 - d1
            self.mydims = (mydims0,) + self.dims[1:]
        else:
            self.mydims = ()

        fullshape = self.mydims + self.myshape

        if data is not None:
            if data.shape != fullshape:
                raise ValueError(
                    f'Bad shape for data: {data.shape} != {fullshape}')
            if data.dtype != dtype:
                raise ValueError(
                    f'Bad dtype for data: {data.dtype} != {dtype}')
            if xp is not None:
                assert (xp is np) == isinstance(
                    data, (np.ndarray, NDArrayReader)), xp
        else:
            data = (xp or np).empty(fullshape, dtype)

        self.data = data
        self.xp: ModuleType
        if isinstance(data, (np.ndarray, NDArrayReader)):
            self.xp = np
        else:
            from gpaw.gpu import cupy as cp
            self.xp = cp
        self._matrix: Matrix | None = None

    def new(self, data=None) -> DistributedArrays:
        raise NotImplementedError

    def copy(self):
        return self.new(data=self.data.copy())

    def __getitem__(self, index):
        raise NotImplementedError

    def __iter__(self):
        for index in range(self.dims[0]):
            yield self[index]

    def flat(self):
        if self.dims == ():
            yield self
        else:
            for index in np.indices(self.dims).reshape((len(self.dims), -1)).T:
                yield self[tuple(index)]

    def to_xp(self, xp):
        if xp is self.xp:
            # Disable assert for now as it would fail with our HIP-rfftn hack!
            # assert xp is np, 'cp -> cp should not be needed!'
            return self
        if xp is np:
            return self.new(data=self.xp.asnumpy(self.data))
        else:
            return self.new(data=xp.asarray(self.data))

    @property
    def matrix(self) -> Matrix:
        if self._matrix is not None:
            return self._matrix

        nx = prod(self.myshape)
        shape = (self.dims[0], prod(self.dims[1:]) * nx)
        myshape = (self.mydims[0], prod(self.mydims[1:]) * nx)
        dist = (self.comm, -1, 1)

        data = self.data.reshape(myshape)
        self._matrix = Matrix(*shape, data=data, dist=dist)

        return self._matrix

    def matrix_elements(self,
                        other,
                        *,
                        out: Matrix = None,
                        symmetric: bool = None,
                        function=None,
                        domain_sum=True,
                        cc: bool = False) -> Matrix:
        if symmetric is None:
            symmetric = self is other
        if function:
            other = function(other)

        M1 = self.matrix
        M2 = other.matrix
        out = M1.multiply(M2, opb='C', alpha=self.dv,
                          symmetric=symmetric, out=out)
        if not cc:
            out.complex_conjugate()
        # operate_and_multiply(self, self.layout.dv, out, function, ...)

        self._matrix_elements_correction(M1, M2, out, symmetric)

        if domain_sum:
            self.domain_comm.sum(out.data)

        return out

    def _matrix_elements_correction(self,
                                    M1: Matrix,
                                    M2: Matrix,
                                    out: Matrix,
                                    symmetric: bool) -> None:
        """Hook for PlaneWaveExpansion."""
        pass

    def abs_square(self,
                   weights: Array1D,
                   out: UniformGridFunctions) -> None:
        """Add weighted absolute square of data to output array.

        See also :xkcd:`849`.
        """
        raise NotImplementedError

    def gather(self, out=None, broadcast=False):
        raise NotImplementedError

    def interpolate(self,
                    plan1: fftw.FFTPlans = None,
                    plan2: fftw.FFTPlans = None,
                    grid: UniformGrid = None,
                    out: UniformGridFunctions = None) -> UniformGridFunctions:
        raise NotImplementedError


def operate_and_multiply(psit1, dv, out, operator, psit2):
    if psit1.comm:
        if psit2 is not None:
            assert psit2.comm is psit1.comm
        if psit1.comm.size > 1:
            out.comm = psit1.comm
            out.state = 'a sum is needed'

    comm = psit1.matrix.dist.comm
    N = len(psit1)
    n = (N + comm.size - 1) // comm.size
    mynbands = len(psit1.matrix.array)

    buf1 = psit1.new(nbands=n, dist=None)
    buf2 = psit1.new(nbands=n, dist=None)
    half = comm.size // 2
    psit = psit1.view(0, mynbands)
    if psit2 is not None:
        psit2 = psit2.view(0, mynbands)

    for r in range(half + 1):
        rrequest = None
        srequest = None

        if r < half:
            srank = (comm.rank + r + 1) % comm.size
            rrank = (comm.rank - r - 1) % comm.size
            skip = (comm.size % 2 == 0 and r == half - 1)
            n1 = min(rrank * n, N)
            n2 = min(n1 + n, N)
            if not (skip and comm.rank < half) and n2 > n1:
                rrequest = comm.receive(buf1.array[:n2 - n1], rrank, 11, False)
            if not (skip and comm.rank >= half) and len(psit1.array) > 0:
                srequest = comm.send(psit1.array, srank, 11, False)

        if r == 0:
            if operator:
                operator(psit1.array, psit2.array)
            else:
                psit2 = psit

        if not (comm.size % 2 == 0 and r == half and comm.rank < half):
            m12 = psit2.matrix_elements(psit, symmetric=(r == 0), cc=True,
                                        serial=True)
            n1 = min(((comm.rank - r) % comm.size) * n, N)
            n2 = min(n1 + n, N)
            out.array[:, n1:n2] = m12.array[:, :n2 - n1]

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        psit = buf1
        buf1, buf2 = buf2, buf1

    requests = []
    blocks = []
    nrows = (comm.size - 1) // 2
    for row in range(nrows):
        for column in range(comm.size - nrows + row, comm.size):
            if comm.rank == row:
                n1 = min(column * n, N)
                n2 = min(n1 + n, N)
                if mynbands > 0 and n2 > n1:
                    requests.append(
                        comm.send(out.array[:, n1:n2].T.conj().copy(),
                                  column, 12, False))
            elif comm.rank == column:
                n1 = min(row * n, N)
                n2 = min(n1 + n, N)
                if mynbands > 0 and n2 > n1:
                    block = np.empty((mynbands, n2 - n1), out.dtype)
                    blocks.append((n1, n2, block))
                    requests.append(comm.receive(block, row, 12, False))

    comm.waitall(requests)
    for n1, n2, block in blocks:
        out.array[:, n1:n2] = block



# {   -1,   -1}:  On entry to PDLASRT parameter number 9 had an illegal value

# works with 'sl_default': (2, 2, 32)

import pytest
from ase.build import fcc100, add_adsorbate
from gpaw import GPAW, ConvergenceError
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl


@pytest.mark.skip(reason='TODO')
def test_scalapack_pdlasrt_hang():
    assert world.size == 4

    slab = fcc100('Cu', size=(2, 2, 2))
    add_adsorbate(slab, 'O', 1.1, 'hollow')
    slab.center(vacuum=3.0, axis=2)

    if compiled_with_sl():
        parallel = {'domain': (1, 1, 4), 'sl_default': (2, 2, 64)}
    else:
        parallel = None

    calc = GPAW(mode='lcao',
                kpts=(2, 2, 1),
                txt='-',
                maxiter=1,
                parallel=parallel)

    slab.calc = calc
    try:
        slab.get_potential_energy()
    except ConvergenceError:
        pass



import numpy as np
from gpaw.utilities.blas import gemmdot
from gpaw.berryphase import get_overlap


class Wannier90:
    def __init__(self, calc, seed=None, bands=None, orbitals_ai=None,
                 spin=0):

        if seed is None:
            seed = calc.atoms.get_chemical_formula()
        self.seed = seed

        if bands is None:
            bands = range(calc.get_number_of_bands())
        self.bands = bands

        Na = len(calc.atoms)
        if orbitals_ai is None:
            orbitals_ai = []
            for ia in range(Na):
                ni = 0
                setup = calc.wfs.setups[ia]
                for l, n in zip(setup.l_j, setup.n_j):
                    if not n == -1:
                        ni += 2 * l + 1
                orbitals_ai.append(range(ni))

        self.calc = calc
        self.bands = bands
        self.Nn = len(bands)
        self.Na = Na
        self.orbitals_ai = orbitals_ai
        self.Nw = np.sum([len(orbitals_ai[ai]) for ai in range(Na)])
        self.kpts_kc = calc.get_ibz_k_points()
        self.Nk = len(self.kpts_kc)
        self.spin = spin


def write_input(calc,
                seed=None,
                bands=None,
                orbitals_ai=None,
                mp=None,
                plot=False,
                num_iter=100,
                write_xyz=False,
                write_rmn=False,
                translate_home_cell=False,
                dis_num_iter=200,
                dis_froz_max=0.1,
                dis_mix_ratio=0.5,
                search_shells=None,
                spinors=False):

    if seed is None:
        seed = calc.atoms.get_chemical_formula()

    if bands is None:
        bands = range(calc.get_number_of_bands())

    Na = len(calc.atoms)
    if orbitals_ai is None:
        orbitals_ai = []
        for ia in range(Na):
            ni = 0
            setup = calc.wfs.setups[ia]
            for l, n in zip(setup.l_j, setup.n_j):
                if not n == -1:
                    ni += 2 * l + 1
            orbitals_ai.append(range(ni))
    assert len(orbitals_ai) == Na

    Nw = np.sum([len(orbitals_ai[ai]) for ai in range(Na)])
    if spinors:
        Nw *= 2
        new_bands = []
        for n in bands:
            new_bands.append(2 * n)
            new_bands.append(2 * n + 1)
        bands = new_bands

    f = open(seed + '.win', 'w')

    pos_ac = calc.spos_ac
    # pos_av = calc.atoms.get_positions()
    # cell_cv = calc.atoms.get_cell()
    # pos_ac = np.dot(pos_av, np.linalg.inv(cell_cv))

    print('begin projections', file=f)
    for ia, orbitals_i in enumerate(orbitals_ai):
        setup = calc.wfs.setups[ia]
        l_i = []
        n_i = []
        for n, l in zip(setup.n_j, setup.l_j):
            if not n == -1:
                l_i += (2 * l + 1) * [l]
                n_i += (2 * l + 1) * [n]
        r_c = pos_ac[ia]
        for orb in orbitals_i:
            l = l_i[orb]
            n = n_i[orb]
            print('f=%1.2f, %1.2f, %1.2f : s ' % (r_c[0], r_c[1], r_c[2]),
                  end='', file=f)
            print('# n = %s, l = %s' % (n, l), file=f)

    print('end projections', file=f)
    print(file=f)

    if spinors:
        print('spinors = True', file=f)
    else:
        print('spinors = False', file=f)
    print('write_hr = True', file=f)
    if write_xyz:
        print('write_xyz = True', file=f)
    if write_rmn:
        print('write_tb = True', file=f)
        print('write_rmn = True', file=f)
    if translate_home_cell:
        print('translate_home_cell = True', file=f)
    print(file=f)
    print('num_bands       = %d' % len(bands), file=f)

    if search_shells is not None:
        print("search_shells = {0}".format(search_shells), file=f)

    maxn = max(bands)
    if maxn + 1 != len(bands):
        diffn = maxn - len(bands)
        print('exclude_bands : ', end='', file=f)
        counter = 0
        for n in range(maxn):
            if n not in bands:
                counter += 1
                if counter != diffn + 1:
                    print('%d,' % (n + 1), sep='', end='', file=f)
                else:
                    print('%d' % (n + 1), file=f)
    print(file=f)

    print('guiding_centres = True', file=f)
    print('num_wann        = %d' % Nw, file=f)
    print('num_iter        = %d' % num_iter, file=f)
    print(file=f)

    if len(bands) > Nw:
        ef = calc.get_fermi_level()
        print('fermi_energy  = %2.3f' % ef, file=f)
        print('dis_froz_max  = %2.3f' % (ef + dis_froz_max), file=f)
        print('dis_num_iter  = %d' % dis_num_iter, file=f)
        print('dis_mix_ratio = %1.1f' % dis_mix_ratio, file=f)
    print(file=f)

    print('begin unit_cell_cart', file=f)
    for cell_c in calc.atoms.cell:
        print('%14.10f %14.10f %14.10f' % (cell_c[0], cell_c[1], cell_c[2]),
              file=f)
    print('end unit_cell_cart', file=f)
    print(file=f)

    print('begin atoms_frac', file=f)
    for atom, pos_c in zip(calc.atoms, pos_ac):
        print(atom.symbol, end='', file=f)
        print('%14.10f %14.10f %14.10f' % (pos_c[0], pos_c[1], pos_c[2]),
              file=f)
    print('end atoms_frac', file=f)
    print(file=f)

    if plot:
        print('wannier_plot   = True', file=f)
        print('wvfn_formatted = True', file=f)
        print(file=f)

    if mp is not None:
        N_c = mp
    else:
        N_c = calc.wfs.kd.N_c
    print('mp_grid =', N_c[0], N_c[1], N_c[2], file=f)
    print(file=f)
    print('begin kpoints', file=f)

    for kpt in calc.get_bz_k_points():
        print('%14.10f %14.10f %14.10f' % (kpt[0], kpt[1], kpt[2]), file=f)
    print('end kpoints', file=f)

    f.close()


def write_projections(calc, seed=None, spin=0, orbitals_ai=None, soc=None):

    if seed is None:
        seed = calc.atoms.get_chemical_formula()

    bands = get_bands(seed)
    Nn = len(bands)

    spinors = False

    win_file = open(seed + '.win')
    for line in win_file.readlines():
        l_e = line.split()
        if len(l_e) > 0:
            if l_e[0] == 'spinors':
                spinors = l_e[2]
                if spinors in ['T', 'true', '1', 'True']:
                    spinors = True
                else:
                    spinors = False
            if l_e[0] == 'num_wann':
                Nw = int(l_e[2])
            if l_e[0] == 'mp_grid':
                Nk = int(l_e[2]) * int(l_e[3]) * int(l_e[4])
                assert Nk == len(calc.get_bz_k_points())

    Na = len(calc.atoms)
    if orbitals_ai is None:
        orbitals_ai = []
        for ia in range(Na):
            ni = 0
            setup = calc.wfs.setups[ia]
            for l, n in zip(setup.l_j, setup.n_j):
                if not n == -1:
                    ni += 2 * l + 1
            orbitals_ai.append(range(ni))
    assert len(orbitals_ai) == Na

    if spinors:
        new_orbitals_ai = []
        for orbitals_i in orbitals_ai:
            new_orbitals_i = []
            for i in orbitals_i:
                new_orbitals_i.append(2 * i)
                new_orbitals_i.append(2 * i + 1)
            new_orbitals_ai.append(new_orbitals_i)
        orbitals_ai = new_orbitals_ai

    Ni = 0
    for orbitals_i in orbitals_ai:
        Ni += len(orbitals_i)
    assert Nw == Ni

    f = open(seed + '.amn', 'w')

    print('Kohn-Sham input generated from GPAW calculation', file=f)
    print('%10d %6d %6d' % (Nn, Nk, Nw), file=f)

    P_kni = np.zeros((Nk, Nn, Nw), complex)
    for ik in range(Nk):
        if spinors:
            P_ani = soc[ik].P_amj
        else:
            P_ani = calc.wfs.kpt_qs[ik][spin].P_ani
        for i in range(Nw):
            icount = 0
            for ai in range(Na):
                ni = len(orbitals_ai[ai])
                P_ni = P_ani[ai][bands]
                P_ni = P_ni[:, orbitals_ai[ai]]
                P_kni[ik, :, icount:ni + icount] = P_ni.conj()
                icount += ni

    for ik in range(Nk):
        for i in range(Nw):
            for n in range(Nn):
                P = P_kni[ik, n, i]
                data = (n + 1, i + 1, ik + 1, P.real, P.imag)
                print('%4d %4d %4d %18.12f %20.12f' % data, file=f)

    f.close()


def write_eigenvalues(calc, seed=None, spin=0, soc=None):

    if seed is None:
        seed = calc.atoms.get_chemical_formula()

    bands = get_bands(seed)

    f = open(seed + '.eig', 'w')

    for ik in range(len(calc.get_bz_k_points())):
        if soc is None:
            e_n = calc.get_eigenvalues(kpt=ik, spin=spin)
        else:
            e_n = soc[ik].eig_m
        for i, n in enumerate(bands):
            data = (i + 1, ik + 1, e_n[n])
            print('%5d %5d %14.6f' % data, file=f)

    f.close()


def write_overlaps(calc, seed=None, spin=0, soc=None, less_memory=False):

    if seed is None:
        seed = calc.atoms.get_chemical_formula()

    if soc is None:
        spinors = False
    else:
        spinors = True

    bands = get_bands(seed)
    Nn = len(bands)
    kpts_kc = calc.get_bz_k_points()
    Nk = len(kpts_kc)

    nnkp = open(seed + '.nnkp', 'r')
    lines = nnkp.readlines()
    for il, line in enumerate(lines):
        if len(line.split()) > 1:
            if line.split()[0] == 'begin' and line.split()[1] == 'nnkpts':
                Nb = eval(lines[il + 1].split()[0])
                i0 = il + 2
                break

    f = open(seed + '.mmn', 'w')

    print('Kohn-Sham input generated from GPAW calculation', file=f)
    print('%10d %6d %6d' % (Nn, Nk, Nb), file=f)

    icell_cv = (2 * np.pi) * np.linalg.inv(calc.wfs.gd.cell_cv).T
    r_g = calc.wfs.gd.get_grid_point_coordinates()
    Ng = np.prod(np.shape(r_g)[1:]) * (spinors + 1)

    dO_aii = []
    for ia in calc.wfs.kpt_u[0].P_ani.keys():
        dO_ii = calc.wfs.setups[ia].dO_ii
        if spinors:
            # Spinor projections require doubling of the (identical) orbitals
            dO_jj = np.zeros((2 * len(dO_ii), 2 * len(dO_ii)), complex)
            dO_jj[::2, ::2] = dO_ii
            dO_jj[1::2, 1::2] = dO_ii
            dO_aii.append(dO_jj)
        else:
            dO_aii.append(dO_ii)

    wfs = calc.wfs

    def wavefunctions(bz_index):
        if spinors:
            # For spinors, G denotes spin and grid: G = (s, gx, gy, gz)
            return soc[bz_index].wavefunctions(
                calc, periodic=True)[bands]
        # For non-spinors, G denotes grid: G = (gx, gy, gz)
        return np.array([wfs.get_wave_function_array(n, bz_index, spin,
                                                     periodic=True)
                         for n in bands])

    if not less_memory:
        u_knG = []
        for ik in range(Nk):
            u_nG = wavefunctions(ik)
            u_knG.append(u_nG)

    P_kani = []
    for ik in range(Nk):
        if spinors:
            P_kani.append(soc[ik].P_amj)
        else:
            P_kani.append(calc.wfs.kpt_qs[ik][spin].P_ani)

    for ik1 in range(Nk):
        if less_memory:
            u1_nG = wavefunctions(ik1)
        else:
            u1_nG = u_knG[ik1]
        for ib in range(Nb):
            # b denotes nearest neighbor k-points
            line = lines[i0 + ik1 * Nb + ib].split()
            ik2 = int(line[1]) - 1
            if less_memory:
                u2_nG = wavefunctions(ik2)
            else:
                u2_nG = u_knG[ik2]

            G_c = np.array([int(line[i]) for i in range(2, 5)])
            bG_v = np.dot(G_c, icell_cv)
            u2_nG = u2_nG * np.exp(-1.0j * gemmdot(bG_v, r_g, beta=0.0))
            bG_c = kpts_kc[ik2] - kpts_kc[ik1] + G_c
            bG_v = np.dot(bG_c, icell_cv)  # Overwrite bG_v
            M_mm = get_overlap(calc,
                               bands,
                               np.reshape(u1_nG, (len(u1_nG), Ng)),
                               np.reshape(u2_nG, (len(u2_nG), Ng)),
                               P_kani[ik1],
                               P_kani[ik2],
                               dO_aii,
                               bG_v)
            indices = (ik1 + 1, ik2 + 1, G_c[0], G_c[1], G_c[2])
            print('%3d %3d %4d %3d %3d' % indices, file=f)
            for m1 in range(len(M_mm)):
                for m2 in range(len(M_mm)):
                    M = M_mm[m2, m1]
                    print('%20.12f %20.12f' % (M.real, M.imag), file=f)

    f.close()


def get_bands(seed):
    win_file = open(seed + '.win')
    exclude_bands = None
    for line in win_file.readlines():
        l_e = line.split()
        if len(l_e) > 0:
            if l_e[0] == 'num_bands':
                Nn = int(l_e[2])
            if l_e[0] == 'exclude_bands':
                exclude_bands = line.split()[2]
                exclude_bands = [int(n) - 1 for n in exclude_bands.split(',')]
    if exclude_bands is None:
        bands = range(Nn)
    else:
        bands = range(Nn + len(exclude_bands))
        bands = [n for n in bands if n not in exclude_bands]
    win_file.close()

    return bands


def write_wavefunctions(calc, soc=None, spin=0, seed=None):

    wfs = calc.wfs

    if soc is None:
        spinors = False
    else:
        spinors = True

    if seed is None:
        seed = calc.atoms.get_chemical_formula()

    bands = get_bands(seed)
    Nn = len(bands)
    Nk = len(calc.get_ibz_k_points())

    for ik in range(Nk):
        if spinors:
            # For spinors, G denotes spin and grid: G = (s, gx, gy, gz)
            u_nG = soc[ik].wavefunctions(calc, periodic=True)
        else:
            # For non-spinors, G denotes grid: G = (gx, gy, gz)
            u_nG = np.array([wfs.get_wave_function_array(n, ik, spin)
                             for n in bands])

        f = open('UNK%s.%d' % (str(ik + 1).zfill(5), spin + 1), 'w')
        grid_v = np.shape(u_nG)[1:]
        print(grid_v[0], grid_v[1], grid_v[2], ik + 1, Nn, file=f)
        for n in range(Nn):
            for iz in range(grid_v[2]):
                for iy in range(grid_v[1]):
                    for ix in range(grid_v[0]):
                        u = u_nG[n, ix, iy, iz]
                        print(u.real, u.imag, file=f)
        f.close()
